#ifndef LLVM_ANALYSIS_CALLSITEANALYSIS_H
#define LLVM_ANALYSIS_CALLSITEANALYSIS_H
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/Instruction.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/PassManager.h"
#include "llvm/IR/Dominators.h"
#include "llvm/Pass.h"
#include <map>
#include <optional>
#include <memory>
#include <set>

struct DevirtFunctionPropertyD;
namespace llvm {
namespace aocc {
struct CallSite {
  CallBase *CB;
  CallSite() :CB(nullptr) { }
  CallSite(CallBase &C) {
    if (isa<InvokeInst>(&C) || isa<CallInst>(&C))
      CB = &C;
    else
      CB = nullptr;
  }
  CallSite(Value *I) {
    if (isa<InvokeInst>(I) || isa<CallInst>(I))
      CB = cast<CallBase>(I);
    else
      CB = nullptr;
  }
  operator bool() const { return (CB != nullptr); }
  Function *getCalledFunction() const { return CB->getCalledFunction(); }
  Value *getCalledValue() { return CB->getCalledValue(); }
  Value *getCalledValue() const { return CB->getCalledValue(); }
  Function *getParentFunction() { return CB->getFunction(); }
  const Function *getParentFunction() const { return CB->getFunction(); }
  Instruction *getInstruction() { return CB; }
  Instruction *getInstruction() const { return CB; }
  CallBase *operator ->() { return CB; }
  CallBase *operator ->() const { return CB; }
  CallingConv::ID getCallingConv() const { return CB->getCallingConv(); }
  Type *getType() const { return CB->getType(); }
  bool arg_empty() const { return CB->arg_empty(); }
  User::op_iterator arg_begin() { return CB->arg_begin(); }
  User::const_op_iterator arg_begin() const { return CB->arg_begin(); }
  User::op_iterator arg_end() { return CB->arg_end(); }
  User::const_op_iterator arg_end() const { return CB->arg_end(); }
  iterator_range<User::op_iterator> args() { return CB->args(); }
};

typedef Constant* Exception;
inline Exception asException(Value *V) { return (dyn_cast<Constant>(V)); }
typedef std::optional<std::set<Exception>> ExceptionSetOpt;
typedef std::optional<std::set<Function *>> FunctionSetOpt;
typedef std::optional<Exception> ExceptionOpt;
typedef std::pair<GlobalVariable *, unsigned long> VTableOffsetPair;
typedef std::optional<std::set<VTableOffsetPair>> GVSetOpt;

struct DevirtFunctionProperty {
  DevirtFunctionPropertyD * DFP = nullptr;
  void run(Module *M, Function *F, DominatorTree *DT);
  std::set<Function *> getPossibleImpl(const CallSite& C);
  std::set<VTableOffsetPair> getPossibleVTables(const CallSite& C);
  ~DevirtFunctionProperty();
};

struct CSFunctionProperty {
  std::shared_ptr<DevirtFunctionProperty> D;
  std::map<Function *, ExceptionSetOpt> EscapingExceptions;
  FunctionSetOpt getPossibleImpl(const CallSite& C);
  GVSetOpt getPossibleVTables(const CallSite& C);
  FunctionSetOpt getPossibleImplWithSig(const CallSite& C);
  ExceptionSetOpt getEscapingExceptions(const CallSite& C);
  ExceptionSetOpt getEscapingExceptions(Function *F);
  bool isAllocFailureException(Function *F);
  bool isExceptionDischarge(BasicBlock *EHCleanup, const Exception& E);
  bool isExceptionRethrow(BasicBlock *EHCleanup, const Exception& E);
  std::optional<int> getBranchForException(Value *C, const Exception& E);
  bool isTrackingStdBadAlloc(const Exception& E);
  void run(Module& M, ModuleAnalysisManager& AM);
  void run(Module& M, Function& F, FunctionAnalysisManager& FAM);
};

class CallSiteAnalysis :public AnalysisInfoMixin<CallSiteAnalysis> {
  friend AnalysisInfoMixin<CallSiteAnalysis>;
  static AnalysisKey Key;
  CSFunctionProperty FP;
public:
  using Result = CSFunctionProperty;
  Result run(Module&, ModuleAnalysisManager& AM);
  Result run(Function&, FunctionAnalysisManager& AM);
};

typedef std::pair<Value *, int> location_t;
struct LocationPool
{
  typedef std::set<int> AccessSet;
  static void merge(AccessSet& one, const AccessSet& other);
  struct Loc {
    std::vector<int> a;
    std::vector<int> i;
    Value *v = nullptr;
    int s;
    enum kind_t { INV, UNK, AREF, CALL, REF, LOC, ARG, GLOB, LOCAL } k;
    bool isc() const { return (k == CALL); }
    bool isr() const { return (k == REF); }
    bool isl() const { return (k == LOC); }
    bool isa() const { return (k == ARG); }
    bool isg() const { return (k == GLOB); }
    bool isu() const { return (k == UNK); }
    bool islv() const { return (k == LOCAL); }
    bool isar() const { return (k == AREF); }
    explicit Loc(Value *v = nullptr, kind_t k = UNK, int s = -1) :v(v), k(k), s(s) { }
    bool operator<(const Loc& O) const {
      if (k != O.k) return (k < O.k);
      if (v != O.v) return (v < O.v);
      if (a != O.a) return (a < O.a);
      if (s != O.s) return (s < O.s);
      return (i < O.i);
    }
    bool operator==(const Loc& O) const { return (v == O.v && i == O.i && k == O.k && a == O.a && s == O.s); }
    void dump() const;
  };
  std::vector<Loc> ll;
  LocationPool *ref = nullptr;
  const Loc& at(int l) const { return (ref ? ref->at(l) : ll.at(l)); }
  Loc& at(int l) { return (ref ? ref->at(l) : ll.at(l)); }
  void put(int i, const Loc& l) { assert(ref == nullptr); ll[i] = l; }
  void setref(LocationPool *nlp) {
    ll.clear();
    ref = nlp;
  }
  int loc(const Loc& );
  int rewrite(int, std::map<int, int>& );
  int ix(const Loc& ) const;
  int ixvk(const Loc& ) const;
  AccessSet loc(Value *);
  AccessSet loc(location_t );
  std::string llstr(int l) const;
  std::string llstr(const AccessSet& ) const;
  void remap(LocationPool& nlp);
  void remap(LocationPool& nlp, std::map<int, int>& imap);
};

struct FunctionInterfaceProperty
{
  Function *F;
  LocationPool lp;
  LocationPool::AccessSet r;
  LocationPool::AccessSet w;
  LocationPool::AccessSet u;
  LocationPool::AccessSet loc(Value *v) { return lp.loc(v); }
  LocationPool::AccessSet loc(location_t v) { return lp.loc(v); }
  int loc(const LocationPool::Loc& l) { return lp.loc(l); }
  const LocationPool::Loc& at(int l) const { return lp.at(l); }
  std::string llstr(int ) const;
  std::string llstr(const LocationPool::AccessSet& ) const;
  bool alias(int, int) const;
  bool rmem(int p) const;
  Type *ty(int p) const;
  bool isua(int p) const;
  void remap(LocationPool& newlp);
  void remap(LocationPool& newlp, std::map<int, int>& remap);
  void subst(int u, const LocationPool& clp);
  void rewrite(std::map<int, int>& imap);
  void inlinec(int u, const FunctionInterfaceProperty& );
  raw_ostream& write(raw_ostream& ) const;
  Value *cvalue(int u) const;
  Function *cfunc(int u) const;
  typedef function_ref<FunctionInterfaceProperty&(Function *)> FIPGetter;
  void inlinec(FIPGetter );
  FunctionInterfaceProperty(Function *F = nullptr) :F(F) { }
};

class FunctionInterfaceAnalysis :public AnalysisInfoMixin<FunctionInterfaceAnalysis> {
  friend AnalysisInfoMixin<FunctionInterfaceAnalysis>;
  static AnalysisKey Key;
public:
  using Result = FunctionInterfaceProperty;
  Result run(Function&, FunctionAnalysisManager& AM);
};

class FunctionInterfaceRun
{
  Function& F;
  FunctionAnalysisManager& AM;
  using Result = FunctionInterfaceAnalysis::Result;
  Result frun(CallBase *CB);
public:
  FunctionInterfaceRun(Function& F, FunctionAnalysisManager& AM) :F(F), AM(AM) { }
  Result itf();
  Result itf(CallBase *);
};

class FunctionInterfacePrinterPass :public PassInfoMixin<FunctionInterfacePrinterPass> {
  raw_ostream& OS;
public:
  explicit FunctionInterfacePrinterPass(raw_ostream &OS);
  PreservedAnalyses run(Function &F, FunctionAnalysisManager &AM);
};

class FunctionInterfaceExpt {
  std::map<Function *, FunctionInterfaceProperty> FM;
  typedef std::pair<CallBase *, Function *> FCKey;
  std::map<FCKey, FunctionInterfaceProperty> FC;
  raw_ostream& OS;
  Module& M;
  ModuleAnalysisManager& AM;
public:
  explicit FunctionInterfaceExpt(raw_ostream& OS, Module& M, ModuleAnalysisManager& AM);
  void ignoreFn(StringRef name);
  void ignoreCppStdFn();
  void ignoreClangStdFn();
  void inlineCall(Function *F, Function *C);
  void inlineCall(Function *F, CallBase *C, Function *CF);
  void dump();
  void dump(StringRef );
  void dump(StringRef, CallBase *);
  void itf(Function *F, CallBase *CF);
  void itf(Function *F);
  FunctionInterfaceProperty& fip(Function *F, CallBase *CF);
  FunctionInterfaceProperty& fip(Function *F);
};

class FunctionInterfaceExptPass :public PassInfoMixin<FunctionInterfaceExptPass> {
  raw_ostream& OS;
public:
  explicit FunctionInterfaceExptPass(raw_ostream& OS) :OS(OS) { }
  PreservedAnalyses run(Module& M, ModuleAnalysisManager& AM);
};

struct IRLocator {
  std::set<Instruction *> l(StringRef s);
  std::set<int> m(StringRef s);
  int m1(StringRef s);
  IRLocator(Module& M, LocationPool& LP) :M(M), LP(LP) { }
  Module& M;
  LocationPool& LP;
};

struct AliasQuery {
  Instruction *ii = nullptr;
  Instruction *jj = nullptr;
  int val = -1;
  enum { REF, MOD, REACH } op;
  // [a,b] means START(a) to END(b).
  // REF: is [val] read in the region [ii,jj].
  // MOD: is [val] written in the region [ii,jj].
  // REACH: does [val] from END(ii) reach BEGIN(jj).
  bool operator<(const AliasQuery& o) const {
    if (op != o.op)
      return (op < o.op);
    else if (ii != o.ii)
      return (ii < o.ii);
    else if (jj != o.jj)
      return (jj < o.jj);
    else
      return (val < o.val);
  }
  static std::set<AliasQuery> qmod(Module& M, LocationPool& LP, StringRef, StringRef );
};
typedef std::set<AliasQuery> AliasQueries;

class CallSiteAliasQuery {
  typedef LocationPool::Loc Loc;
  typedef AliasQuery Q;
public:
  typedef std::map<Q, AliasResult::Kind> Answer;
  typedef std::vector<Q> Queries;
  using Result = Answer;
  CallSiteAliasQuery& q(const Q& qe);
  CallSiteAliasQuery(Module& M, LocationPool& LP, ModuleAnalysisManager& AM) :M(M), LP(LP), AM(AM) { }
  Result run();
private:
  Module& M;
  LocationPool& LP;
  ModuleAnalysisManager& AM;
  Queries qq;
  bool ismod(Instruction *ii, int v);
  bool ismod(Function *CF, int v);
};

class CallSiteAliasPrinterPass :public PassInfoMixin<CallSiteAliasPrinterPass> {
  raw_ostream& OS;
  CallSiteAliasQuery::Queries qq;
  LocationPool LP;
public:
  explicit CallSiteAliasPrinterPass(raw_ostream &OS);
  void qmod(Module&, StringRef, StringRef);
  PreservedAnalyses run(Module& M, ModuleAnalysisManager& AM);
};

#if AOCC_DEVELOPER_BUILD
class CallSiteAliasCPU2017PrinterPass :public PassInfoMixin<CallSiteAliasCPU2017PrinterPass> {
  StringRef AppName;
public:
   CallSiteAliasCPU2017PrinterPass(StringRef AppName);
   PreservedAnalyses run(Module&, ModuleAnalysisManager& );
};

class FunctionInterfaceExptCPU2017Pass :public PassInfoMixin<FunctionInterfaceExptCPU2017Pass> {
  StringRef AppName;
public:
  FunctionInterfaceExptCPU2017Pass(StringRef AppName);
  PreservedAnalyses run(Module&, ModuleAnalysisManager& );
};

#endif
}
}
#endif
