//===---------------- DDGSimilarity.h -----------------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//====  Copyright (c) 2021 Advanced Micro Devices, Inc.  All rights reserved.
//
//               Developed by: Advanced Micro Devices, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// with the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimers.
//
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimers in the documentation
// and/or other materials provided with the distribution.
//
// Neither the names of Advanced Micro Devices, Inc., nor the names of its
// contributors may be used to endorse or promote products derived from this
// Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH
// THE SOFTWARE.
//===----------------------------------------------------------------------===//

#ifndef LLVM_ANALYSIS_DDGSIMILARITY_H
#define LLVM_ANALYSIS_DDGSIMILARITY_H

#include "llvm/IR/PassManager.h"
#include "llvm/Analysis/DDG.h"
#include "llvm/Analysis/DependenceAnalysis.h"
#include <unordered_set>

namespace llvm {

struct pair_hashing {
  template <class T1, class T2>
  std::size_t operator()(std::pair<T1, T2> const &pair) const {
    std::size_t h1 = std::hash<T1>()(pair.first);
    std::size_t h2 = std::hash<T2>()(pair.second);

    return h1 ^ h2;
  }
};

// a node in the product graph
class ProductGraphNode {
public:
  DDGNode *nodeA, *nodeB;
  double similarityValue = 0, updateValue = 0;
  std::vector<ProductGraphNode *> neighbours;

  ProductGraphNode(DDGNode *nodeA, DDGNode *nodeB);

  void addEdge(ProductGraphNode *dst);
  void displayPnodeInfo();
};

// product graph used in similarity flooding
class ProductGraph {
  public:
  std::vector<ProductGraphNode *> nodeList;
  DataDependenceGraph *ddgA, *ddgB;
  std::unordered_map<DDGNode *, int> ddgANodeToIndex, ddgBNodeToIndex;
  std::unordered_map<int, DDGNode *> ddgAIndexToNode, ddgBIndexToNode;
  std::vector<std::vector<double>> similarityMatrix;
  std::unordered_set<std::pair<DDGNode *, DDGNode *>, pair_hashing> edgeSet;

  ProductGraph(DataDependenceGraph *DDGA, DataDependenceGraph *DDGB);
  void printProductGraph();
  void populateSimilarityMatrix();
  double getMaxSimilarity();
  void floodSimilarity();
  double getSimilarityScoreofMatchedNodes(
    std::vector<std::pair<int, int>> &match);
};

class DDGSim {
  public:
  SmallVector<BasicBlock *, 4> clonedBasicBlocks;
  SmallVector<Instruction *, 4> clonedInstructions;
  SmallVector<DataDependenceGraph *, 4> DDGList;

  DDGSim() {}
  void splitIntoWeaklyConnectedComponents(BasicBlock &BB, DependenceInfo &DI,
                                          LLVMContext &context);
};

class DDGSimilarityPass : public PassInfoMixin<DDGSimilarityPass> {
  public:

  DDGSimilarityPass() {}

  PreservedAnalyses run(Module &M, ModuleAnalysisManager &AM);
};
} // namespace llvm

#endif // LLVM_ANALYSIS_DDGSIMILARITY_H
