//===-InefficientAccessCaching.h - Cache inefficient accesses-------------===//
//===---------------------- of Data Structures ---------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_TRANSFORMS_IPO_INEFFICIENTACCESSCACHING_H
#define LLVM_TRANSFORMS_IPO_INEFFICIENTACCESSCACHING_H

#include "llvm/ADT/Statistic.h"
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/AssumptionCache.h"
#include "llvm/Analysis/DomTreeUpdater.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Analysis/LoopNestAnalysis.h"
#include "llvm/Analysis/MemorySSA.h"
#include "llvm/Analysis/MemorySSAUpdater.h"
#include "llvm/Analysis/PointerCacheMissAnalysis.h"
#include "llvm/Analysis/PostDominators.h"
#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/CodeGen/RuntimeLibcalls.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/Dominators.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/PassManager.h"
#include "llvm/InitializePasses.h"
#include "llvm/Pass.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Transforms/IPO.h"
#include "llvm/Transforms/Utils/AnalyzeInterProceduralLoopRegions.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/LCSSA.h"
#include "llvm/Transforms/Utils/LoopSimplify.h"
#include "llvm/Transforms/Utils/LoopSplitUtils.h"
#include "llvm/Transforms/Utils/LoopUtils.h"
#include "llvm/Transforms/Utils/MemoizationUtils.h"

namespace llvm {

struct CacheInfo {

  // Used to store newly created types for aocc_map cache interactions.
  Type *CacheType = nullptr;
  // Map to relate the position of each load instruction to its caching
  // type's position. Used to calculate the GEP's indices for loading/
  // storing to the cache.
  // The second entry in this map determines the shape of subtype in the cache
  // for the respective instruction.
  std::map<Instruction *, SmallVector<unsigned, 2>> CacheTypeIndexMap;
  // The cache pointer after loading the cache through aocc_get_cache.
  Value *AOCCCache = nullptr;
  // The Key for Nth cache, loaded in the header of the outermost loop of the
  // LoopNest.
  Value *NthCache = nullptr;
  // The boolean value of the result for the comparison of NthCacheKey to null.
  Value *IsCached = nullptr;
  // The boolean value to decide whether we can cache an element.
  Value *NeedToCache = nullptr;
  // Auxiliary caching conditions.
  Value *AuxCacheCond = nullptr;

  /// \return the type at the given index in a consolidated type CacheType. For
  /// example if the consolidated type is:
  ///
  /// { ptr, i32, [4 x { ptr, ptr } ] }
  ///
  /// If the index in the given \param Indices is [2,1], then the type at these
  /// indices is calculated by getting the nested types. In the parent type,
  /// the type at index 2 is the array of another consolidated type. The type
  /// at index 1 for this consolidated type is ptr, thus is returned.
  Type *getTypeAt(SmallVector<unsigned, 2> Indices) {
    Type *SubType = CacheType;
    for (unsigned Idx : Indices) {
      unsigned NumSubTypes = SubType->getNumContainedTypes();
      if (Idx >= NumSubTypes)
        return nullptr;
      SubType = SubType->getContainedType(Idx);
      if (SubType->isArrayTy())
        SubType = SubType->getArrayElementType();
    }
    return SubType;
  }
  // struct aoccCacheLibFns {
  //   StringRef Name;
  //   Type *ReturnType;
  //   std::initializer_list<Type*> ArgTypes;
  // };

  /// TODO: These types has to be changed for m32 cases.
  // const struct aoccCacheLibFns aoccCacheLibFnTable[4] = {
  //   { "aocc_cache_count", Type::getInt64Ty(*Ctx), { Type::getInt64PtrTy(*Ctx) } },
  //   { "aocc_init_cache", Type::getVoidTy(*Ctx), { Type::getInt64PtrTy(*Ctx), Type::getInt64Ty(*Ctx), Type::getInt64Ty(*Ctx) } },
  //   { "aocc_get_cache", PointerType::getUnqual(CacheType), { Type::getInt64PtrTy(*Ctx) } },
  //   { "aocc_clear_cache", Type::getVoidTy(*Ctx), {} }
  // };
};

class IACaching {
public:
  using ReuseLoopMap =
      std::map<Loop *,
               SmallVector<pair<Instruction *, SmallVector<Value *, 8>>, 8>>;

private:
  Module *M = nullptr;
  Function *F = nullptr;
  LoopInfo *LI = nullptr;
  Loop *ReuseLoop = nullptr;
  Loop *EncapsulatingLoop = nullptr;
  Value *EncapLoopIndVar = nullptr;
  ScalarEvolution *SE = nullptr;
  DominatorTree *DT = nullptr;
  PostDominatorTree *PDT = nullptr;
  AAResults *AA = nullptr;
  AssumptionCache *AC = nullptr;
  MemorySSAUpdater *MSSAU = nullptr;
  OptimizationRemarkEmitter *ORE = nullptr;

  SmallVector<SmallVector<Instruction *>> LoadChains;
  ReuseLoopMap CallsToCache;
  SmallVector<BranchInst *> TargetExceptionBranches;
  // This vector holds all the load instructions that are found to be
  // inefficient accesses inside a LoopNest.
  SmallVector<Instruction *, 2> TargetInsts;
  ValueToValueMapTy RegionCloneVMap;
  SmallVector<ValueToValueMapTy *> SubLoopVMaps;
  // Value to value map for the cloned innermost loop.
  ValueToValueMapTy InnerLoopVMap;
  SmallVector<BasicBlock *> ClonedRegionBlocks;
  // The global stateVariable copied from IACachingPass class.
  GlobalVariable *stateVar = nullptr;
  // The root value being loaded first in the loop nest in the chain of
  // inefficient access loads.
  // Instruction *RootOfChain = nullptr;
  std::map<unsigned, std::pair<Loop *, SmallVector<Instruction *, 2>>>
      LoadGroups;
  std::map<Loop *, Value *> LoopToIndvarMap;

  CacheInfo CI;

  /// Splits the Outermost loop in two loops with iteration range split like:
  /// First loop = [OrigIVStart, CachingSize]
  /// Second loop = [CachingSize + 1, OrigIVEnd]
  bool splitLoop();

  /// Store inefficient loads in TargetInsts.
  void getInefficientLoads();

  /// Store inefficient calls in TargetInsts.
  void getInefficientCalls();

  /// Return true if the instruction \param I is the LoopNest \param LN
  /// variant.
  bool isLoopNestVariant(Instruction *, LoopNest *);

  /// Collect the loop and indvar info for the given loads.
  void collectLoopNestInfo();

  /// Construct the caching type required for generating loads/stores to the
  /// cache for enefficient loads. \return false if not able to do so.
  bool constructCachingTypes();

  /// \return the root load instruction in the loop nest of the chain of
  /// inefficient loads. The root of chain is the first load in the chain with
  /// induction variable of outermost loop as index, in the access pattern for
  /// an inefficient access of load.
  Instruction *getRootOfChain();

  /// Try to group the inefficient loads according to the induction variable as
  /// their index and store them in \param LoadGroups. Each map entry in the
  /// parameter is a group.
  void tryToConsolidate(
      std::map<unsigned, std::pair<Loop *, SmallVector<Instruction *, 2>>> &);

  void performRegionCloning();

  void cloneRegionsForInnerLoops(SmallVector<Loop *> &);

  void createCacheUpdateCond();

  /// Replace the inefficient loads in the loop nest by cloning the innermost
  /// loop and then adding the loads/stores at the appropriate locations with
  /// respective conditions.
  void replaceInefficientLoads();

  /// Initialize the AOCC cache's map interface, its accesses and each entry's
  /// presence condition accordingly. These initializations are made as
  /// follows:
  ///
  /// At LoopNest's outermost loop's preheader:
  /// 1. Check if cache created/present for this object by calling the AOCC map
  ///    function 'aocc_cache_count'. Cache is present if it returns true.
  /// 2. If the cache is not present, initialize the cache for this object by
  ///    calling the AOCC map function 'aocc_init_cache'.
  /// 3. If cache is initialized already or just initialized, get the cache
  ///    by calling AOCC map function 'aocc_get_cache'.
  ///
  /// At LoopNest's outermost loop's header:
  /// 1. Create the GEP for loading the IndVar'th element from the cache.
  /// 2. Create the GEP and load the key (i.e. FieldValueMap pointer in the
  ///    case of xalanc) from IndVar'th element of cache.
  /// 3. Compare the above key with nullptr to check if IndVar'th element is
  ///    present in the cache or not.
  /// 4. Also create the condition for checking if the loaded element should
  ///    be cached or not by creating AND condition of (if the element is not
  ///    present in the cache) and (if innermost loop bound is <=
  ///    MaxCacheSubtypeArraySize).
  void initializeCache();

  /// Clone the innermost loop \param InnermostLoop with the proper structure
  /// around it.
  Loop *cloneInnermostLoop(Loop *);

  /// For single exit loops/multi exit loops, get the latch exit block for loop
  /// \param Loop.
  BasicBlock *getLatchExitBlock(Loop *);

  /// Calculate insert points for the stores of inefficient access loads to
  /// the cache. Store these insert points by mapping each inefficient load
  /// to the list of its insert points in \param InsertPoints.
  void
  getInsertPoints(std::map<Instruction *, SmallVector<Instruction *, 2>> &);

  /// Map the store insert points \param InsertPoints to the insert points for
  /// inserting the loads in \param MappedInsertPoints in the original
  /// innermost loop.
  void
  mapInsertPoints(std::map<Instruction *, SmallVector<Instruction *, 2>> &,
                  std::map<Instruction *, SmallVector<Instruction *, 2>> &);

  /// Get the reverse mapped value for the value \param Key in the value to
  /// value map for the cloned innermost loop.
  const Value *getOriginalValueFromVMap(Value *);

  /// Find a region/splice in the function which holds the following
  /// properties:
  /// 1. All the values defined in the region have uses only inside the region.
  /// 2. All the target basic blocks for jump instructions are in the region.
  /// 3. No instruction can potentially write to memory.
  /// These properties ensure that there will be no illegalities if a region of
  /// code is sinked inside an if condition.
  ///
  /// The End of the regions is the insert points given in the
  /// \param InsertPoints. To find the start of the regions, we iterate each
  /// instruction upwards until they satisfy the aforementioned properties.
  /// For each load instruction, the map entry in \param RegionStartMapper will
  /// have an entry pair as region start, region end.
  void analyzeFunctionForCodeSplice(
      std::map<Instruction *, SmallVector<Instruction *, 2>> &,
      std::map<Instruction *, std::map<Instruction *, Instruction *>> &);

  /// \return the previous instruction of the instruction \param Key. The
  /// \param BBList is the list of all basic blocks in the function in post
  /// order.
  Instruction *getPrevInstruction(Instruction *,
                                  SmallVector<BasicBlock *, 8> &);

  /// \return true if an instruction \param I is valid in the splice. An
  /// instruction is valid if it holds the properties mentioned in the comments
  /// of function analyzeFunctionForCodeSplice. \param SpliceEnd is the
  /// instruction where the splice ends.
  bool isValidInSplice(Instruction *, Instruction *);

  /// Remove exception handling code in the caching path if possible. Since we
  /// are RAUWing the inefficient loads in the caching path and loading the
  /// values from cache, some of the original exceptions can be removed.
  void removePossibleExceptions();

  /// Calculate and store the branch instructions in \param ExceptionGens which
  /// are branching to an exception handling basic block.
  void getExceptionTargetBranches(SmallVector<BranchInst *, 2> &);

  /// Create GEP indices vector in \param GEPIndicesVals for the load/store of
  /// instruction \param I for the type \param MyType. These GEP indices are
  /// provided by the indices in \param GEPIndices according to their position
  /// in \param MyType.
  void createGEPIndicesFromType(Instruction *, Type *,
                                const SmallVector<unsigned, 2> &,
                                SmallVector<Value *, 2> &);

public:
  IACaching(Module *M_, Function *F_,
            SmallVector<SmallVector<Instruction *>> *LC, LoopInfo *LI_,
            ScalarEvolution *SE_, DominatorTree *DT_, PostDominatorTree *PDT_,
            AAResults *AA_, AssumptionCache *AC_, MemorySSAUpdater *MSSAU_,
            OptimizationRemarkEmitter *ORE_, GlobalVariable *stateVar_)
      : M(M_), F(F_), LI(LI_), SE(SE_), DT(DT_), PDT(PDT_), AA(AA_), AC(AC_),
        MSSAU(MSSAU_), ORE(ORE_), stateVar(stateVar_) {
    TargetInsts.clear();
    CallsToCache.clear();
    LoadChains = *LC;
    ReuseLoop = nullptr;
    getInefficientLoads();
    /// NOTE: Need to recalculate SCEVs, SE is not consistent before and after
    // the analysis.
    SE->forgetAllLoops();
  }

  IACaching(Module *M_, Function *F_, ReuseLoopMap *CTC, LoopInfo *LI_,
            ScalarEvolution *SE_, DominatorTree *DT_, PostDominatorTree *PDT_,
            AAResults *AA_, AssumptionCache *AC_, MemorySSAUpdater *MSSAU_,
            OptimizationRemarkEmitter *ORE_, GlobalVariable *stateVar_)
      : M(M_), F(F_), LI(LI_), SE(SE_), DT(DT_), PDT(PDT_), AA(AA_), AC(AC_),
        MSSAU(MSSAU_), ORE(ORE_), stateVar(stateVar_) {
    TargetInsts.clear();
    CallsToCache = *CTC;
    LoadChains.clear();
    getInefficientCalls();
    /// NOTE: Need to recalculate SCEVs, SE is not consistent before and after
    // the analysis.
    SE->forgetAllLoops();
  }

  ~IACaching() {
    for (auto VMap : SubLoopVMaps)
      delete (VMap);
  }

  bool runImpl();

  // Run legality checks for the structure of the IR. Analytical legality
  // checks are performed by some of th analyses.
  bool runLegalityChecks();
};

/// Pass to run the IACaching transformation optimization.
class IACachingPass : public PassInfoMixin<IACachingPass> {
  GlobalVariable *stateVar = nullptr;

public:
  PreservedAnalyses run(Module &M, ModuleAnalysisManager &);

  // We need to run the transformation on the cloned function. If the runtime
  // check for stateVariable is LOOPING or CACHING, we call the transformed
  // function, otherwise we call the original function. Since the
  // transformation runs on the cloned function, we clone the function and add
  // runtime checks here itself.
  Function *cloneFunctionWithRuntimeChecks(Module &, ModuleAnalysisManager &,
                                           Function &, ValueToValueMapTy *,
                                           std::vector<Loop *> &LoopRegions,
                                           AnalyzeInterProceduralLoopRegions &AIP,
                                           const std::vector<llvm::Function*>&);

  // Function find interprocedural loop and call graph regions.
  std::vector<Loop *> findInterProceduralLoopRegions(Module &M,
    ModuleAnalysisManager &AM, Function &F,
    const std::vector<Function *> &Constructors);

  // Create the global stateVariable used to set and determine the state in
  // which the current transformation is in. These states are DEFAULT,
  // LOOPING, CACHING and INVALID.
  GlobalVariable *createGlobalStateVariable(Module *);

  // Returns the global stateVariable.
  GlobalVariable *getGlobalStateVariable() { return stateVar; }

  std::map<Function *, SmallVector<Instruction *>>
  getFuncCallstoCacheMap(Module &, ModuleAnalysisManager &);

  void removeAnalysisFailedChains(CXXMemoizationAnalysis *,
                                  SmallVector<SmallVector<Instruction *>> &);
};

} // namespace llvm

#endif // LLVM_TRANSFORMS_IPO_INEFFICIENTACCESSCACHING_H
