//===-- Scalar.h - Scalar Transformations -----------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
// Modifications Copyright (c) 2019 Advanced Micro Devices, Inc. All rights reserved.
// Notified per clause 4(b) of the license.
//
//===----------------------------------------------------------------------===//
//               Copyright (c) 2015 Advanced Micro Devices, Inc.
//               All rights reserved.
//               Developed by: Advanced Micro Devices, Inc.
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// with the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is furnished
// to do so, subject to the following conditions:
// Redistributions of source code must retain the above copyright notice, this list
// of conditions and the following disclaimers.
// Redistributions in binary form must reproduce the above copyright notice, this
// list of conditions and the following disclaimers in the documentation and/or
// other materials provided with the distribution.
// Neither the names of Advanced Micro Devices, Inc., nor the names of its
// contributors may be used to endorse or promote products derived from this Software
// without specific prior written permission.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
// OR OTHER DEALINGS WITH THE SOFTWARE.
//===----------------------------------------------------------------------===//
//
// This header file defines prototypes for accessor functions that expose passes
// in the Scalar transformations library.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_TRANSFORMS_SCALAR_H
#define LLVM_TRANSFORMS_SCALAR_H

#include "llvm/Transforms/Utils/SimplifyCFGOptions.h"
#include <functional>

namespace llvm {

class Function;
class FunctionPass;
class ModulePass;
class Pass;

//===----------------------------------------------------------------------===//
//
// RedundantDbgInstElimination - This pass removes redundant dbg intrinsics
// without modifying the CFG of the function.  It is a FunctionPass.
//
Pass *createRedundantDbgInstEliminationPass();

//===----------------------------------------------------------------------===//
//
// DeadCodeElimination - This pass is more powerful than DeadInstElimination,
// because it is worklist driven that can potentially revisit instructions when
// their other instructions become dead, to eliminate chains of dead
// computations.
//
FunctionPass *createDeadCodeEliminationPass();


//===----------------------------------------------------------------------===//
//
// AddInlineAttributes - Analyze functions and add inline attributes.
//
FunctionPass *createAddInlineAttributesPass();

// AOCC Begin
//===----------------------------------------------------------------------===//
//
// createDefineMemSetPass - This pass removes call to memset lib and defines
// memset locally
FunctionPass *createDefineMemSetPass(int OptLevel = 0);

//===----------------------------------------------------------------------===//
//
// createUndefLocalizationPass - This pass inserts undef initialization of
// local variables so as to take care of SROA and mem2reg transformations
FunctionPass *createUndefLocalizationPass();


//===----------------------------------------------------------------------===//
//
// InsertMemset  - A pass the replaces the call to f90_mzero8 with memset
//
FunctionPass *createInsertMemset();

// AOCC End

//===----------------------------------------------------------------------===//
//
// GuardWidening - An optimization over the @llvm.experimental.guard intrinsic
// that (optimistically) combines multiple guards into one to have fewer checks
// at runtime.
//
FunctionPass *createGuardWideningPass();


//===----------------------------------------------------------------------===//
//
// LoopGuardWidening - Analogous to the GuardWidening pass, but restricted to a
// single loop at a time for use within a LoopPassManager.  Desired effect is
// to widen guards into preheader or a single guard within loop if that's not
// possible.
//
Pass *createLoopGuardWideningPass();


//===----------------------------------------------------------------------===//
//
// SROA - Replace aggregates or pieces of aggregates with scalar SSA values.
//
FunctionPass *createSROAPass(bool PreserveCFG = true);

//===----------------------------------------------------------------------===//
//
// IndirectCallPromotion - Promote indirect calls to direct calls by replacing
// with if-else checks before the call.
//
Pass *createIndirectCallPromotionPass();

//===----------------------------------------------------------------------===//
//
// Combine instructions into Intrinsics
//
Pass *createPreSLPInstCombinePass();

//===----------------------------------------------------------------------===//
//
// LICM - This pass is a loop invariant code motion and memory promotion pass.
//
Pass *createLICMPass(unsigned OptLevel = 1);

//===----------------------------------------------------------------------===//
//
// LoopVersioningLICM - This pass is a loop multi versioning pass for LICM.
//
Pass *createLoopVersioningLICMPass();

//===----------------------------------------------------------------------===//
//
// JumpLocality - Improve locality for jump targets.
//
Pass *createJumpLocalityLegacyPass();

// LoopSink - This pass sinks invariants from preheader to loop body where
// frequency is lower than loop preheader.
//
Pass *createLoopSinkPass();

//===----------------------------------------------------------------------===//
//
// LoopPredication - This pass does loop predication on guards.
//
Pass *createLoopPredicationPass();

//===----------------------------------------------------------------------===//
//
// LoopStrengthReduce - This pass is strength reduces GEP instructions that use
// a loop's canonical induction variable as one of their indices.
//
Pass *createLoopStrengthReducePass();

//===----------------------------------------------------------------------===//
//
// LoopUnswitch - This pass is a simple loop unswitching pass.
//
Pass *createLoopUnswitchPass(bool OptimizeForSize = false,
                             bool hasBranchDivergence = false);

//===----------------------------------------------------------------------===//
//
// LoopInstSimplify - This pass simplifies instructions in a loop's body.
//
Pass *createLoopInstSimplifyPass();

//===----------------------------------------------------------------------===//
//
// LoopUnroll - This pass is a simple loop unrolling pass.
//
Pass *createLoopUnrollPass(int OptLevel = 2, bool OnlyWhenForced = false,
                           // AOCC params begin
                           bool UnrollAggressive = false,
                           bool SLPUnroll = false,
                           // AOCC params end
                           bool ForgetAllSCEV = false, int Threshold = -1,
                           int Count = -1, int AllowPartial = -1,
                           int Runtime = -1, int UpperBound = -1,
                           int AllowPeeling = -1);

//===----------------------------------------------------------------------===//
//
//
// LoopReversal.
//
Pass *createLoopReversalPass();

//===----------------------------------------------------------------------===//
//
//
// LoopFirstContinue.
//
Pass* createLoopFirstContinuePass();

//===----------------------------------------------------------------------===//
//
// LoopRotate - This pass is a simple loop rotating pass.
//
Pass *createLoopRotatePass(int MaxHeaderSize = -1, bool PrepareForLTO = false);

//===----------------------------------------------------------------------===//
//
// PromoteMemoryToRegister - This pass is used to promote memory references to
// be register references. A simple example of the transformation performed by
// this pass is:
//
//        FROM CODE                           TO CODE
//   %X = alloca i32, i32 1                 ret i32 42
//   store i32 42, i32 *%X
//   %Y = load i32* %X
//   ret i32 %Y
//
//FunctionPass *createPromoteMemoryToRegisterPass();

//===----------------------------------------------------------------------===//
//
// DemoteRegisterToMemoryPass - This pass is used to demote registers to memory
// references. In basically undoes the PromoteMemoryToRegister pass to make cfg
// hacking easier.
//
FunctionPass *createDemoteRegisterToMemoryPass();
extern char &DemoteRegisterToMemoryID;

//===----------------------------------------------------------------------===//
//
// Reassociate - This pass reassociates commutative expressions in an order that
// is designed to promote better constant propagation, GCSE, LICM, PRE...
//
// For example:  4 + (x + 5)  ->  x + (4 + 5)
//
FunctionPass *createReassociatePass();

//===----------------------------------------------------------------------===//
//
// CFGSimplification - Merge basic blocks, eliminate unreachable blocks,
// simplify terminator instructions, convert switches to lookup tables, etc.
//
FunctionPass *createCFGSimplificationPass(
    SimplifyCFGOptions Options = SimplifyCFGOptions(),
    std::function<bool(const Function &)> Ftor = nullptr);

//===----------------------------------------------------------------------===//
//
// FlattenCFG - flatten CFG, reduce number of conditional branches by using
// parallel-and and parallel-or mode, etc...
//
FunctionPass *createFlattenCFGPass();

//===----------------------------------------------------------------------===//
//
// CFG Structurization - Remove irreducible control flow
//
///
/// When \p SkipUniformRegions is true the structizer will not structurize
/// regions that only contain uniform branches.
Pass *createStructurizeCFGPass(bool SkipUniformRegions = false);

//===----------------------------------------------------------------------===//
//
// BreakConstantGEPs - Change constant GEPs to GEP expressions.
//
FunctionPass *createBreakConstantGEPsPass();

//===----------------------------------------------------------------------===//
//
// BreakCriticalEdges - Break all of the critical edges in the CFG by inserting
// a dummy basic block. This pass may be "required" by passes that cannot deal
// with critical edges. For this usage, a pass must call:
//
//   AU.addRequiredID(BreakCriticalEdgesID);
//
// This pass obviously invalidates the CFG, but can update forward dominator
// (set, immediate dominators, tree, and frontier) information.
//
FunctionPass *createBreakCriticalEdgesPass();
extern char &BreakCriticalEdgesID;

  FunctionPass *createBlockReorderingPass(bool AssumeClosed = false);
//===----------------------------------------------------------------------===//
//
// LoopSimplify - Insert Pre-header blocks into the CFG for every function in
// the module.  This pass updates dominator information, loop information, and
// does not add critical edges to the CFG.
//
//   AU.addRequiredID(LoopSimplifyID);
//
Pass *createLoopSimplifyPass();
extern char &LoopSimplifyID;

// begin AOCC
Pass *createPreRvTransformsPass();
extern char &PreRvTransformsID;
// end AOCC
//===----------------------------------------------------------------------===//
//
// TailCallElimination - This pass eliminates call instructions to the current
// function which occur immediately before return instructions.
//
FunctionPass *createTailCallEliminationPass(bool EnableInlineRecursionConstProp = false);

//===----------------------------------------------------------------------===//
//
// EarlyCSE - This pass performs a simple and fast CSE pass over the dominator
// tree.
//
FunctionPass *createEarlyCSEPass(bool UseMemorySSA = false);

//===----------------------------------------------------------------------===//
//
// MergedLoadStoreMotion - This pass merges loads and stores in diamonds. Loads
// are hoisted into the header, while stores sink into the footer.
//
FunctionPass *createMergedLoadStoreMotionPass(bool SplitFooterBB = false);

//===----------------------------------------------------------------------===//
//
// ConstantHoisting - This pass prepares a function for expensive constants.
//
FunctionPass *createConstantHoistingPass();

//===----------------------------------------------------------------------===//
//
// Sink - Code Sinking
//
FunctionPass *createSinkingPass();

//===----------------------------------------------------------------------===//
//
// LowerAtomic - Lower atomic intrinsics to non-atomic form
//
Pass *createLowerAtomicPass();

//===----------------------------------------------------------------------===//
//
// LowerGuardIntrinsic - Lower guard intrinsics to normal control flow.
//
Pass *createLowerGuardIntrinsicPass();

//===----------------------------------------------------------------------===//
//
// LowerWidenableCondition - Lower widenable condition to i1 true.
//
Pass *createLowerWidenableConditionPass();

//===----------------------------------------------------------------------===//
//
// MergeICmps - Merge integer comparison chains into a memcmp
//
Pass *createMergeICmpsLegacyPass();

//===----------------------------------------------------------------------===//
//
// InferAddressSpaces - Modify users of addrspacecast instructions with values
// in the source address space if using the destination address space is slower
// on the target. If AddressSpace is left to its default value, it will be
// obtained from the TargetTransformInfo.
//
FunctionPass *createInferAddressSpacesPass(unsigned AddressSpace = ~0u);
extern char &InferAddressSpacesID;

//===----------------------------------------------------------------------===//
//
// LowerExpectIntrinsics - Removes llvm.expect intrinsics and creates
// "block_weights" metadata.
FunctionPass *createLowerExpectIntrinsicPass();

//===----------------------------------------------------------------------===//
//
// TLSVariableHoist - This pass reduce duplicated TLS address call.
//
FunctionPass *createTLSVariableHoistPass();

//===----------------------------------------------------------------------===//
//
// LowerConstantIntrinsicss - Expand any remaining llvm.objectsize and
// llvm.is.constant intrinsic calls, even for the unknown cases.
//
FunctionPass *createLowerConstantIntrinsicsPass();

//===----------------------------------------------------------------------===//
//
// PartiallyInlineLibCalls - Tries to inline the fast path of library
// calls such as sqrt.
//
FunctionPass *createPartiallyInlineLibCallsPass();

//===----------------------------------------------------------------------===//
//
// SimplifyPowLibCalls - Tries to simplify the pow library call to simpler
// equivalent instructions.
//
FunctionPass *createSimplifyPowLibCallsPass();

//===----------------------------------------------------------------------===//
//
// ScalarizerPass - Converts vector operations into scalar operations
//
FunctionPass *createScalarizerPass();

//===----------------------------------------------------------------------===//
//
// SeparateConstOffsetFromGEP - Split GEPs for better CSE
//
FunctionPass *createSeparateConstOffsetFromGEPPass(bool LowerGEP = false);

//===----------------------------------------------------------------------===//
//
// SpeculativeExecution - Aggressively hoist instructions to enable
// speculative execution on targets where branches are expensive.
//
FunctionPass *createSpeculativeExecutionPass();

// Same as createSpeculativeExecutionPass, but does nothing unless
// TargetTransformInfo::hasBranchDivergence() is true.
FunctionPass *createSpeculativeExecutionIfHasBranchDivergencePass();

//===----------------------------------------------------------------------===//
//
// EliminateSimilarExpPass - Remove the similar add/sub expressions from loop,
// whose scev exp is not same but almost similar and can be eliminated by
// applying the adjustment
FunctionPass *createEliminateSimilarExprPass();

//===----------------------------------------------------------------------===//
//
FunctionPass *
createIVSplittingPass();

//===---------------------------------------------------------------------===//
//
// PromoteConstantToVariable - Move constants to variable
// It combines the same constant to variable
FunctionPass *
createPromoteConstantToVariablePass();

//===----------------------------------------------------------------------===//
//
// StraightLineStrengthReduce - This pass strength-reduces some certain
// instruction patterns in straight-line code.
//
FunctionPass *createStraightLineStrengthReducePass();



//===----------------------------------------------------------------------===//
//
// StructArrayCopy - Region Based Struct Array Copy- In case of array
// of structures, if a loop accesses only some fields, copy the fields
// temporarily to another array (if nested loop).
//
FunctionPass *createStructArrayCopyPass();

//===----------------------------------------------------------------------===//
//
// Initial code:
// call llvm.masked.gather(..., <compile-time unknown mask>, ...)
//
// TO
//
// Cond: br  <mask all true>, If, Else
// If:
//  call llvm.masked.gather(..., <all-true>, ...)
// Else:
//  call llvm.masked.gather(..., <compile-time unknown mask>, ...)
ModulePass *createScalarizeMaskedGatherLegacyPass(); //AOCC

//===----------------------------------------------------------------------===//
//

//===----------------------------------------------------------------------===//
//

FunctionPass *createFlangPropertyPass(int mode = 0); // AOCC
FunctionPass *createFlangPropertyCleanupPass(); // AOCC

// NaryReassociate - Simplify n-ary operations by reassociation.
//
FunctionPass *createNaryReassociatePass();

//===----------------------------------------------------------------------===//
//

FunctionPass *createLoopDistributeAdvancedPass();

FunctionPass *createIfToForPass(bool ProcessAllLoopsByDefault);

//===----------------------------------------------------------------------===//
//
// LoopUnrotate - Unrotate loops.
//
FunctionPass *createLoopUnrotatePass();

//===----------------------------------------------------------------------===//
//
// LoopBlock - Block loops.
//
FunctionPass *createLoopBlockPass();


//===----------------------------------------------------------------------===//
//
// PeepholeLoopSimplify - Simplify loop by peephole optimizations.
//
FunctionPass *createPeepholeLoopSimplifyPass();

//===----------------------------------------------------------------------===//
//
// RemoveTripCountIV - Remove IV generated from flang to track trip count.
//
FunctionPass *createRemoveTripCountIVPass();

// LoopSimplifyCFG - This pass performs basic CFG simplification on loops,
// primarily to help other loop passes.
//
Pass *createLoopSimplifyCFGPass();


// Begin AOCC.
//===----------------------------------------------------------------------===//
//
//// ConvertNTStorePass - generate nontemporal stores.
////
//
FunctionPass *createConvertNTStorePass(int Level);
FunctionPass *createAdjustAlignmentForVectorNTStoresPass();
FunctionPass *createPropagateTBAAPassPass();
extern char &PropagateTBAAPassID;
// end AOCC.

// Begin AOCC.
//===----------------------------------------------------------------------===//
//
//// PeelLoopsForNTStoreVectorization - peel loops for nontemporal store
//// vectorization.
////
FunctionPass *createPeelLoopsForNTStoreVectorizationPass();
// end AOCC.

//===----------------------------------------------------------------------===//
//
// LoopDataPrefetch - Perform data prefetching in loops.
//
FunctionPass *createLoopDataPrefetchPass();
//===----------------------------------------------------------------------===//
//
// LoopBoundCache - prune iterations for inner most loop guarded by loop indices
//
FunctionPass *createLoopIterPruningPass();

//===----------------------------------------------------------------------===//
//
FunctionPass *createDynamicCastSimplifyPass();

// LoopSimplifyCFG - This pass performs basic CFG simplification on loops,
// primarily to help other loop passes.
//
Pass *createLoopSimplifyCFGPass();

//===----------------------------------------------------------------------===//
//
FunctionPass *createSparseArrayIteration(); // AOCC
FunctionPass *createSparseLoadElimination(); //AOCC
FunctionPass *createSparseGuardIntoIteration(); // AOCC
FunctionPass *createBreakReductionSLP(); // AOCC
FunctionPass *createDoubleToIntConvert(); // AOCC
FunctionPass *createMoveLoadSliceGSLP(); // AOCC

//===----------------------------------------------------------------------===//
//
// StripMining - Strip Mining.
//
FunctionPass *createStripMiningPass();

//===----------------------------------------------------------------------===//
//
// LoopFuse - Fuse loops.
//
FunctionPass *createLoopFuseAOCCPass();

// AOCC Begin
//===----------------------------------------------------------------------===/
//
// WholeLoopNestRTAliasChecksPass - This pass performs the alias checks outside
// loop nests if possible and marks the loop ready for optimizations.
//
FunctionPass *createWholeLoopNestRTAliasChecksPass();
//AOCC End


//===----------------------------------------------------------------------===//
//
// This pass does instruction simplification on each
// instruction in a function.
//
FunctionPass *createInstSimplifyLegacyPass();

//===----------------------------------------------------------------------===//
//
// PreCodeGenInstLoweringLegacy
//
FunctionPass *createPreCodeGenInstLoweringLegacy();

//===----------------------------------------------------------------------===//
//
// PropHigherOrderMask
//
FunctionPass *createPropHigherOrderMaskLegacy();

//===----------------------------------------------------------------------===//
//
// createScalarizeMaskedMemIntrinPass - Replace masked load, store, gather
// and scatter intrinsics with scalar code when target doesn't support them.
//
FunctionPass *createScalarizeMaskedMemIntrinLegacyPass();
FunctionPass *createFlowMarkerFunctionPass(const char *); // AOCC
FunctionPass *createFlowMarkerFunctionPass(const char *, const char *); // AOCC
} // End llvm namespace

#endif
