//===- llvm/Transfors/Scalar/LoopFission/FPG.h ------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
///
/// \file
/// This file defines the Loop Fission Partition Graph (FPG).
///
//===----------------------------------------------------------------------===//

#ifndef LLVM_TRANSFORMS_SCALAR_FPG_H
#define LLVM_TRANSFORMS_SCALAR_FPG_H

#include "llvm/ADT/DirectedGraph.h"
#include "llvm/ADT/SmallSet.h"
#include "llvm/ADT/SparseBitVector.h"
#include "llvm/Analysis/DDG.h"
#include "llvm/Analysis/PostDominators.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/LoopUtils.h"
#include "llvm/Transforms/Utils/LoopVersioning.h"
#include "llvm/Transforms/Utils/ValueMapper.h"

namespace llvm {

class FIGNode;
class FIGEdge;
using FIGNodeBase = DGNode<FIGNode, FIGEdge>;
using FIGEdgeBase = DGEdge<FIGNode, FIGEdge>;
using FIGBase = DirectedGraph<FIGNode, FIGEdge>;

/// Fission Interference Graph Node
/// The graph can represent the following types of nodes:
/// 1. Default node containing one or more DDG node.
/// 2. Root Node is a special node that connects to all components such that
///    there is always a path from it to any node in the graph.
class FIGNode : public FIGNodeBase {
public:
  using DDGNodeSet = SmallPtrSet<DDGNode *, 4>;
  using InstructionListType = DDGNode::InstructionListType;

  enum class NodeKind { Default, Root };

  enum class ScalarExpandKind { SE_Expand, SE_Use, SE_Control, SE_None };

  FIGNode(const FIGNode &N) = delete;
  FIGNode(FIGNode &&N)
      : FIGNodeBase(std::move(N)), DDGNodes(std::move(N.DDGNodes)),
        Kind(N.getKind()) {}
  virtual ~FIGNode();

  // Create a node containing the given DDG node \p DDGN.
  static FIGNode *create(DDGNode &DDGN);

  FIGNode &operator=(const FIGNode &N) {
    DGNode::operator=(N);
    DDGNodes = N.DDGNodes;
    Kind = N.Kind;
    return *this;
  }

  FIGNode &operator=(FIGNode &&N) {
    DGNode::operator=(std::move(N));
    DDGNodes = std::move(N.DDGNodes);
    Kind = N.Kind;
    return *this;
  }

  const DDGNodeSet &getNodes() const { return DDGNodes; }
  NodeKind getKind() const { return Kind; }

  /// Insert the given DDG nodes \p Nodes into this FIG node.
  void insertNodes(const DDGNodeSet &Nodes) {
    DDGNodes.insert(Nodes.begin(), Nodes.end());
  }
  void markMergedSCCNode() { MergedSCCNode = true; }
  void markLoopFIGNode() { LoopFIGNode = true; }
  bool isLoopFIGNode() const { return LoopFIGNode; }

  /// Return true if this FIG node contains the given DDG node \p DDGN.
  bool containsNode(const DDGNode &DDGN) const {
    return llvm::find(DDGNodes, &DDGN) != DDGNodes.end();
  }
  /// Populate \p IList with the instructions contained in the DDG nodes
  /// associated with this node. Return true if at least one instructions was
  /// collected, and false otherwise.
  bool collectInstructions(InstructionListType &IList) const;

  /// Populate \p IList with the instructions contained in the DDG nodes
  /// associated with this node, for which the predicate function \p Pred
  /// evaluates to true. Return true if at least one instructions was collected,
  /// and false otherwise.
  bool collectInstructions(InstructionListType &IList,
                           function_ref<bool(Instruction *)> Pred) const;

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FIGNode *N) {
    return N->getKind() == NodeKind::Default;
  }

  /// Return true if DDG node \p N1 is connected to \p N2 via a path
  /// containing only DDG edges of the given kind \p EK.
  static bool areConnectedByPath(const DDGNode &N1, const DDGNode &N2,
                                 DDGEdge::EdgeKind EK);

  /// Collect the DDG nodes that have an incident DDG edge of kind \p EK
  /// stemming from any of the DDG nodes associated with FIG node \p N. Do not
  /// consider input dependencies if \p SkipInputDependencies is true.
  static bool collectTargetNodes(const FIGNode &N, DDGEdge::EdgeKind EK,
                                 bool SkipInputDependencies,
                                 const DataDependenceGraph &DDG,
                                 DDGNodeSet &Targets);

  bool hasPI() const {
    if (MergedSCCNode)
      return true;
    return any_of(getNodes(), [&](const DDGNode *Src) {
      if (isa<PiBlockDDGNode>(Src))
        return true;
      return false;
    });
  }

  bool containsOnlyStoreNodes() const {
    if (hasPI()) {
      auto MemoryInsts = [](Instruction *I) {
        return I->mayReadOrWriteMemory();
      };
      SmallVector<Instruction *, 4> NodeIList;
      if (collectInstructions(NodeIList, MemoryInsts)) {
        for (auto *I : NodeIList) {
          if (!isa<StoreInst>(I))
            return false;
        }
        return true;
      }
    }
    return false;
  }

  bool isValidMemoryPI() const {
    auto MemoryInsts = [](Instruction *I) { return I->mayReadOrWriteMemory(); };
    if (hasPI()) {
      SmallVector<Instruction *, 4> NodeIList;
      if (collectInstructions(NodeIList, MemoryInsts))
        return true;
    }
    return false;
  }

  void setScalarExpand(bool V = true) {
    if (V)
      SEKind = ScalarExpandKind::SE_Expand;
    else
      SEKind = ScalarExpandKind::SE_None;
  }
  bool canScalarExpand() const {
    return (SEKind == ScalarExpandKind::SE_Expand);
  }
  void setScalarExpandUse( bool V = true ) { 
    if (V) SEKind = ScalarExpandKind::SE_Use; 
    else SEKind = ScalarExpandKind::SE_None; 
  }
  bool hasScalarExpandedUse() const {
    return SEKind == ScalarExpandKind::SE_Use;
  }
  void setControlExpandNode() { SEKind = ScalarExpandKind::SE_Control; }
  bool isControlExpand() const {
    return SEKind == ScalarExpandKind::SE_Control;
  }

  void dump() const;

protected:
  FIGNode(const NodeKind Kind) : FIGNodeBase(), Kind(Kind) {}

private:
  FIGNode(DDGNode &N) { DDGNodes.insert(&N); }

  /// Data dependence graph nodes associated with this FIG node.
  DDGNodeSet DDGNodes;

  bool LoopFIGNode = false;
  bool MergedSCCNode = false;

  ScalarExpandKind SEKind = ScalarExpandKind::SE_None;

  /// The node kind.
  NodeKind Kind = NodeKind::Default;
};

/// Subclass of FIGNode representing the root node of the fission interference
/// graph. There should only be one such node in a given graph.
class RootFIGNode final : public FIGNode {
public:
  RootFIGNode() : FIGNode(NodeKind::Root) {}
  RootFIGNode(const RootFIGNode &) = delete;
  RootFIGNode(RootFIGNode &&N) : FIGNode(std::move(N)) {}

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FIGNode *N) {
    return N->getKind() == NodeKind::Root;
  }
  static bool classof(const RootFIGNode *N) { return true; }
};

/// Fission Interefence Graph Edge.
/// Represents an edge between 2 nodes in the fission interference graph.
/// Edges are used to express data dependencies between nodes and to indicate
/// affinity between the nodes.
/// A root edge connects the root node to one of the components of the fission
/// interference graph.
class FIGEdge : public FIGEdgeBase {
public:
  enum class EdgeKind {
    DefUse,   /// edge represents a def-use dependence
    Memory,   /// edge represents a data dependence
    Affinity, /// edge represents the desire to keep nodes in the same loop
    Control,  /// edge represents control dependence between nodes
    Root      /// edge stems from the graph's root node
  };

  FIGEdge(const FIGEdge &) = delete;
  FIGEdge(FIGEdge &&E) : FIGEdgeBase(std::move(E)), Kind(E.getKind()) {}
  virtual ~FIGEdge() = 0;

  /// Create an edge incident to FIG node \p N with the given edge kind \p EK.
  /// When building an affinity edge its weight is given by \p Weight.
  static FIGEdge *create(FIGNode &N, EdgeKind EK,
                         std::optional<unsigned> Weigth = std::nullopt);

  FIGEdge &operator=(const FIGEdge &E) {
    FIGEdgeBase::operator=(E);
    Kind = E.Kind;
    return *this;
  }

  FIGEdge &operator=(FIGEdge &&E) {
    FIGEdgeBase::operator=(std::move(E));
    Kind = E.Kind;
    return *this;
  }

  EdgeKind getKind() const { return Kind; }

  void dump() const;

protected:
  FIGEdge(FIGNode &N, EdgeKind Kind) : FIGEdgeBase(N), Kind(Kind) {}

private:
  EdgeKind Kind;
};

template <FIGEdge::EdgeKind EK> class FIGEdgeOfKind final : public FIGEdge {
public:
  FIGEdgeOfKind(FIGNode &N) : FIGEdge(N, EK) {}
  FIGEdgeOfKind(const FIGEdgeOfKind &) = delete;
  FIGEdgeOfKind(FIGEdgeOfKind &&E) : FIGEdge(std::move(E)) {}

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FIGEdge *N) { return N->getKind() == EK; }
  static bool classof(const FIGEdgeOfKind *) { return true; }
};

using DefUseFIGEdge = FIGEdgeOfKind<FIGEdge::EdgeKind::DefUse>;
using MemoryFIGEdge = FIGEdgeOfKind<FIGEdge::EdgeKind::Memory>;
using RootFIGEdge = FIGEdgeOfKind<FIGEdge::EdgeKind::Root>;
using ControlFIGEdge = FIGEdgeOfKind<FIGEdge::EdgeKind::Control>;

/// A weighted edge used to represents affinity between nodes. The greater the
/// weight the larger the affinity between the nodes connected by the edge.
class AffinityFIGEdge final : public FIGEdge {
public:
  AffinityFIGEdge(FIGNode &N, int Weight)
      : FIGEdge(N, EdgeKind::Affinity), Weight(Weight) {
    assert(Weight != 0 && "Must have non zero weight");
  }
  AffinityFIGEdge(const AffinityFIGEdge &) = delete;
  AffinityFIGEdge(AffinityFIGEdge &&E)
      : FIGEdge(std::move(E)), Weight(E.getWeight()) {}

  AffinityFIGEdge &operator=(const AffinityFIGEdge &E) {
    FIGEdge::operator=(E);
    Weight = E.Weight;
    return *this;
  }

  AffinityFIGEdge &operator=(AffinityFIGEdge &&E) {
    FIGEdge::operator=(std::move(E));
    Weight = E.Weight;
    return *this;
  }

  int getWeight() const { return Weight; };
  void updateWeight(int W) { Weight += W; }

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FIGEdge *N) {
    return N->getKind() == EdgeKind::Affinity;
  }
  static bool classof(const AffinityFIGEdge *N) { return true; }

private:
  int Weight;
};

class NodeConnector;

/// Fission Interference Graph
/// A directed acyclic graph where each node represents a set of ddg nodes
///  The edges in the graph represent memory, defuse and control dependencies.
class FissionInterferenceGraph final : public FIGBase {
  friend class NodeConnector; // needs to add edges to the graph (to connect
                              // nodes).

public:
  /// Map from a node to its index in the graph, and its reverse map.
  using NodeToIndexMapTy = DenseMap<FIGNode *, unsigned>;
  using IndexToNodeMapTy = DenseMap<unsigned, FIGNode *>;

  using NodeToPartitionsTy = DenseMap<FIGNode *, DenseSet<unsigned>>;
  using PartitionToNodesTy = DenseMap<unsigned, SmallPtrSet<FIGNode *, 4>>;
  using NodeType = FIGNode;
  using NodeListType = SmallVector<NodeType *, 4>;
  using ControlDepCacheTy = DenseMap<BasicBlock *, SmallVector<BasicBlock *,32> >;

  using InstToNodeIdTy = DenseMap<Instruction *, unsigned>;
  using NodeToParentMapTy = DenseMap<unsigned, DenseSet<unsigned>>;

  FissionInterferenceGraph() = delete;
  FissionInterferenceGraph(const FissionInterferenceGraph &) = delete;
  ~FissionInterferenceGraph();

  /// Construct a fission interference graph for the loop nest rooted by \p
  /// OutermostLoop, given the data dependence graph \p DDG.
  static FissionInterferenceGraph *
  create(const DataDependenceGraph &DDG, Loop *OutermostLoop, DominatorTree *DT,
         PostDominatorTree *PDT, ScalarEvolution *SE, LoopInfo *LI);

  bool hasRootNode() const { return (RootNode != nullptr); }

  bool isRootNode(const FIGNode *N) const {
    return (N && isa<RootFIGNode>(*N) && hasRootNode() && *N == getRootNode());
  }

  FIGNode &getRootNode() const {
    assert(RootNode && "Root node is not available yet. Graph construction may "
                       "still be in progress\n");
    return *RootNode;
  }
  DenseSet<unsigned> getParents(unsigned ix) { 
    assert(NodeToParentMap.find(ix) != NodeToParentMap.end());
    return NodeToParentMap[ix];
  }

  const DirectedGraph::NodeListTy &getNodes() const { return Nodes; }
  const DataDependenceGraph &getDDG() const { return DDG; }

  /// Returns the number of nodes in the graph (excluding the root node).
  unsigned size() const {
    return RootNode ? FIGBase::size() - 1 : FIGBase::size();
  }

  /// Retrieve the node \p N graph index.
  unsigned getNodeIndex(const FIGNode &N) const {
    assert(NodeToIndexMap.find(&N) != NodeToIndexMap.end() &&
           "Could not find node in map");
    return NodeToIndexMap.lookup(&N);
  };

  /// Retrieve a graph node given its graph index \p I.
  FIGNode &getNode(unsigned I) const {
    assert(IndexToNodeMap.find(I) != IndexToNodeMap.end() &&
           "Could not find node in map");
    return *IndexToNodeMap.lookup(I);
  };

  /// Return true if any of the DDG nodes contained in FIG node \p N reach any
  /// of the DDG nodes contained in FIG node \p Other via a path containing only
  /// DDG edges of the given kind \p EK.
  static bool areConnectedByPath(const FIGNode &N, const FIGNode &Other,
                                 DDGEdge::EdgeKind EK);

  /// Return true if FIG node \N has a memory dependence with FIG node \p Other.
  /// Do not consider input dependencies if \p SkipInputDependencies is true.
  static bool haveMemoryDependence(const FIGNode &N, const FIGNode &Other,
                                   const DataDependenceGraph &DDG,
                                   bool SkipInputDependencies = false);

  bool haveControlDep(const FIGNode &N1, const FIGNode &N2);

  bool CanIgnoreMemEdge(FIGNode *Node, FIGEdge *E);
  /// Return true FIG node \p N has a def-use dependence with \p Other.
  static bool haveDefUseDependence(const FIGNode &N, const FIGNode &Other) {
    return areConnectedByPath(N, Other, DDGEdge::EdgeKind::RegisterDefUse);
  }

  void createPartitions(NodeToPartitionsTy &NodeToPartitionMap,
                        PartitionToNodesTy &PartitionToNodesMap,
                        bool &invalidFPG, bool WithScalarExpansion);
  bool shouldMergePartitions(NodeToPartitionsTy &NodeToPartitionMap,
                             PartitionToNodesTy &PartitionToNodesMap, unsigned,
                             unsigned, bool);
  void mergePartitions(NodeToPartitionsTy &NodeToPartitionMap,
                       PartitionToNodesTy &PartitionToNodesMap, unsigned,
                       unsigned);

  bool isLoopFIGNode(const FIGNode *N) const;

  void getLoopsFor(const FIGNode *N, SmallPtrSet<Loop *, 4> &Loops,
                   bool InductionOnly) const;

  void dump() const;

  bool hasSCCs() const;
  bool MergeSCCs();
  void viewFIG(const char *suffix = "");
  bool isValid(const Loop *L) const;

  DenseMap<Value *, Value *> &getScalarExpandMap() {
    return ScalarExpandMallocMap;
  }
  bool getControlNodes(FIGNode *N, SmallVectorImpl<FIGNode *> &ControlNodes);

private:
  FissionInterferenceGraph(const DataDependenceGraph &DDG, const Loop *L,
                           DominatorTree *DT, PostDominatorTree *PDT,
                           ScalarEvolution *SE, LoopInfo *LI)
      : FIGBase(), DDG(DDG), L(L), DT(DT), PDT(PDT), SE(SE), LI(LI) {}

  /// This member function uses the Data dependence graph (DDG) to populate the
  /// interference graph. Data dependencies are enforced by connecting the graph
  /// nodes with edges. After construction the interference graph is guaranteed
  /// to be acyclic.
  void populate();

  /// Add the given node \p N to the graph.
  bool addNode(FIGNode &N);

  void DFSUtil(FIGNode &Node, SmallVectorImpl<unsigned> &Visited,
               NodeToPartitionsTy &NodeToPartitionMap,
               PartitionToNodesTy &PartitionToNodesMap,
               SmallVectorImpl<FIGNode *> &Stack, bool WithScalarExpansion);

  /// Connect the \p Src and \p Dst nodes using the given edge \p E.
  /// Return true if the edge was added to the graph and false otherwise (an
  /// existing edge was reused).
  bool addEdge(FIGNode &Src, FIGNode &Dst, FIGEdge &E);

  /// Connect graph nodes with an edge that represents the data dependencies
  /// between their DDG nodes. Return true if any edge was created and false
  /// otherwise.
  bool createDependencyEdges();

  /// Create a root node that connects to every subgraph. For example given:
  ///   G: {N1-->N2,  N3-->N4,  N5}
  /// The root node would be connected to the subgraphs rooted by N1, N3 and N5.
  void createAndConnectRootNode();

  void fixControlEdgesFromLoop();

  void DFStraverse(FIGNode *S, FIGNode *V);

  void updateReachability();

  bool isReachableFrom(FIGNode *, FIGNode *);

  bool findAndMarkScalarExpansionNodes();

private:
  const DataDependenceGraph &DDG;

  const Loop *L = nullptr;

  /// The graph root node (connected to all subgraphs in the graph).
  RootFIGNode *RootNode = nullptr;

  /// A map from a node to its unique graph index.
  NodeToIndexMapTy NodeToIndexMap;

  /// A map from the node graph index to the node.
  IndexToNodeMapTy IndexToNodeMap;

  DenseMap<const Loop *, FIGNode *> LoopToLoopFIGNodeMap;

  DenseSet<unsigned> PINodes;

  DenseMap<FIGNode *, SmallVector<bool, 4>> ReachabilityMatrix;

  NodeToParentMapTy NodeToParentMap;

  NodeToParentMapTy NodeToControlParentMap;

  ControlDepCacheTy ControlDepCache;

  InstToNodeIdTy InstToNodeIdMap;

  DenseMap<Value *, Value *> ScalarExpandMallocMap;

  /// Used to connect nodes in the graph.
  NodeConnector *NC = nullptr;

  DominatorTree *DT;
  PostDominatorTree *PDT;
  ScalarEvolution *SE;
  LoopInfo *LI;
  unsigned DotFileCount = 0;
};

/// A class used to connect nodes in a fission interference graph.
class NodeConnector final {
  friend class FissionInterferenceGraph;
  FissionInterferenceGraph &G;

public:
  NodeConnector() = delete;
  NodeConnector(const NodeConnector &) = delete;
  NodeConnector &operator=(const NodeConnector &) = delete;

  /// Connect nodes for which the given predicate function \p Selector evaluates
  /// to true. The edges created are given the kind \p EK.
  bool
  connectNodes(function_ref<bool(const FIGNode &, const FIGNode &)> Selector,
               FIGEdge::EdgeKind EK);

  /// Connect nodes for which the given predicate function \p Selector evaluates
  /// to true. The affinity edges created are given the weight returned by the
  /// \p GetWeight function.
  bool
  connectNodes(function_ref<bool(const FIGNode &, const FIGNode &)> Selector,
               function_ref<int(const FIGNode &, const FIGNode &)> GetWeight);

  /// Connect nodes for which the given function \p Selector return non-zero
  /// with an affinity edge having weight equal to the returned value of
  /// \p Selector.
  bool connectNodes(
      function_ref<int(const FIGNode &N1, const FIGNode &N2)> Selector);

  /// Connect nodes \p N1 and \p N2 with a new edge of the same kind as the
  /// given edge \p E.
  bool connectNodes(FIGNode &N1, FIGNode &N2, const FIGEdge &E) const;

private:
  NodeConnector(FissionInterferenceGraph &G) : G(G) {}

  /// Connect nodes \p N1 and \p N2 with an edge of the given the kind \p EK.
  bool connectNodes(FIGNode &N1, FIGNode &N2, FIGEdge::EdgeKind EK) const;

  /// Connect nodes \p N1 and \p N2 with an affinity edge of the weight \p
  /// Weight.
  bool connectNodes(FIGNode &N1, FIGNode &N2, int Weight) const;
};

raw_ostream &operator<<(raw_ostream &, const FIGNode &);
raw_ostream &operator<<(raw_ostream &, const FIGEdge &);
raw_ostream &operator<<(raw_ostream &, const FissionInterferenceGraph &);

//===--------------------------------------------------------------------===//
// GraphTraits specializations for the interference graph
//===--------------------------------------------------------------------===//

/// non-const versions of the grapth trait specializations the FIG.
template <> struct GraphTraits<FIGNode *> {
  using NodeRef = FIGNode *;

  static NodeRef FIGGetTargetNode(DGEdge<FIGNode, FIGEdge> *P) {
    return &P->getTargetNode();
  }

  // Provide a mapped iterator so that the GraphTrait-based implementations can
  // find the target nodes without having to explicitly go through the edges.
  using ChildIteratorType =
      mapped_iterator<FIGNode::iterator, decltype(&FIGGetTargetNode)>;
  using ChildEdgeIteratorType = FIGNode::iterator;

  static NodeRef getEntryNode(NodeRef N) { return N; }
  static ChildIteratorType child_begin(NodeRef N) {
    return ChildIteratorType(N->begin(), &FIGGetTargetNode);
  }
  static ChildIteratorType child_end(NodeRef N) {
    return ChildIteratorType(N->end(), &FIGGetTargetNode);
  }
  static ChildEdgeIteratorType child_edge_begin(NodeRef N) {
    return N->begin();
  }
  static ChildEdgeIteratorType child_edge_end(NodeRef N) { return N->end(); }
};

template <>
struct GraphTraits<FissionInterferenceGraph *> : public GraphTraits<FIGNode *> {
  using NodeRef = FIGNode *;
  using nodes_iterator = FissionInterferenceGraph::iterator;

  static NodeRef getEntryNode(FissionInterferenceGraph *FIG) {
    return &FIG->getRootNode();
  }
  static nodes_iterator nodes_begin(FissionInterferenceGraph *FIG) {
    return FIG->findNode(FIG->getRootNode());
  }
  static nodes_iterator nodes_end(FissionInterferenceGraph *FIG) {
    return FIG->end();
  }
};

/// const versions of the grapth trait specializations for the FIG.
template <> struct GraphTraits<const FIGNode *> {
  using NodeRef = const FIGNode *;

  static NodeRef FIGGetTargetNode(const DGEdge<FIGNode, FIGEdge> *P) {
    return &P->getTargetNode();
  }

  // Provide a mapped iterator so that the GraphTrait-based implementations can
  // find the target nodes without having to explicitly go through the edges.
  using ChildIteratorType =
      mapped_iterator<FIGNode::const_iterator, decltype(&FIGGetTargetNode)>;
  using ChildEdgeIteratorType = FIGNode::const_iterator;

  static NodeRef getEntryNode(NodeRef N) { return N; }
  static ChildIteratorType child_begin(NodeRef N) {
    return ChildIteratorType(N->begin(), &FIGGetTargetNode);
  }
  static ChildIteratorType child_end(NodeRef N) {
    return ChildIteratorType(N->end(), &FIGGetTargetNode);
  }
  static ChildEdgeIteratorType child_edge_begin(NodeRef N) {
    return N->begin();
  }
  static ChildEdgeIteratorType child_edge_end(NodeRef N) { return N->end(); }
};

template <>
struct GraphTraits<const FissionInterferenceGraph *>
    : public GraphTraits<const FIGNode *> {
  using NodeRef = const FIGNode *;
  using nodes_iterator = FissionInterferenceGraph::const_iterator;

  static NodeRef getEntryNode(const FissionInterferenceGraph *FIG) {
    return &FIG->getRootNode();
  }
  static nodes_iterator nodes_begin(const FissionInterferenceGraph *FIG) {
    return FIG->findNode(FIG->getRootNode());
  }
  static nodes_iterator nodes_end(const FissionInterferenceGraph *FIG) {
    return FIG->end();
  }
};

class FPGNode;
class FPGEdge;
using FPGNodeBase = DGNode<FPGNode, FPGEdge>;
using FPGEdgeBase = DGEdge<FPGNode, FPGEdge>;

using FPGBase = DirectedGraph<FPGNode, FPGEdge>;

// Data structure to record the info about a node in the
// fisson partition graph. These metrics are used in the heuristics
// to decide if we can merge a partition or not.
struct FPGMetrics {
  unsigned NumStridedStores = 0;
  unsigned NonAffineStores = 0;
  unsigned NumStridedLoads = 0;
  unsigned NonAffineLoads = 0;
  unsigned MaxBitWidth = 0;
  unsigned MaxDepth = 0;
  bool Conditional = false;

  SmallPtrSet<Value *, 4> LdPtrs;
  SmallPtrSet<Value *, 4> StPtrs;

  SmallPtrSet<Instruction *, 4> Stores;
  SmallPtrSet<Instruction *, 4> Loads;

  SmallPtrSet<Instruction *, 4> CIs;
  SmallPtrSet<Instruction *, 4> ComplexOps;
  SmallPtrSet<Instruction *, 4> SharedOps;

  SmallPtrSet<Loop *, 4> Loops;
  void reset() {
    NumStridedStores = 0;
    NonAffineStores = 0;

    NumStridedLoads = 0;
    NonAffineLoads = 0;

    MaxBitWidth = 0;
    Conditional = false;

    StPtrs.clear();
    LdPtrs.clear();
    Stores.clear();
    Loads.clear();
    CIs.clear();
    ComplexOps.clear();
    SharedOps.clear();
    Loops.clear();
  };
  unsigned getNumNonAffineStores() const { return NonAffineStores; }
  unsigned getNumStridedAccesses() const {
    return NumStridedLoads + NumStridedStores;
  }
  void updateStore(Instruction *S, Value *Ptr, bool NonAffine, bool IsStrided) {
    Stores.insert(S);
    StPtrs.insert(Ptr);
    if (NonAffine)
      NonAffineStores++;
    if (IsStrided)
      NumStridedStores++;
  }
  void updateLoad(Instruction *Ld, Value *Ptr, bool NonAffine, bool IsStrided) {
    Loads.insert(Ld);
    LdPtrs.insert(Ptr);
    if (NonAffine)
      NonAffineLoads++;
    if (IsStrided)
      NumStridedLoads++;
  }

  void updateComplexOp(Instruction *C) { ComplexOps.insert(C); }
  void updateFunctionCall(Instruction *F) { CIs.insert(F); }
  void updateSharedOp(Instruction *I) { SharedOps.insert(I); }
  void updateBitWidth(unsigned Bitwidth) { MaxBitWidth = Bitwidth; }
  void setCondtional() { Conditional = true; }

  unsigned getNumStores() const { return Stores.size(); }
  unsigned getNumLoads() const { return Loads.size(); }
  unsigned getNumFunctions() const { return CIs.size(); }
  unsigned getNumComplexOps() const { return ComplexOps.size(); }
  unsigned getNumSharedOps() const { return SharedOps.size(); }
  unsigned getMaxBitWidth() const { return MaxBitWidth; }

  bool isConditional() const { return Conditional; }

  unsigned getMaxLoopDepth() const;

  void print() const;
  bool hasAnyCommonPointers(const FPGMetrics &M, bool Load) const;
  bool hasSharedNodes(const FPGMetrics &M) const;
  bool atSameLoopLevel(const FPGMetrics &M) const;

  bool isTrivial() const {
    return ((getNumStores() == 0) && (getNumLoads() == 0) &&
            (getNumFunctions() == 0) && (getNumComplexOps() == 0));
  }
};

/// A node in the Fission partition Graph. This node represents
/// the set of instructions that can be formed as a loop from the
// original loop nest.
class FPGNode : public FPGNodeBase {
public:
  using FIGNodeSet = SmallPtrSet<FIGNode *, 4>;
  using InstructionListType = FIGNode::InstructionListType;
  using InstSet = SmallPtrSet<Instruction *, 4>;

  enum class NodeKind { Default, Root };

  FPGNode(const FPGNode &N) = delete;
  FPGNode(FPGNode &&N)
      : FPGNodeBase(std::move(N)), FIGNodes(std::move(N.FIGNodes)),
        Kind(N.getKind()) {}
  virtual ~FPGNode();

  // Create a node containing the given DDG node \p DDGN.
  static FPGNode *create(FIGNodeSet &FIGNodes);

  FPGNode &operator=(const FPGNode &N) {
    DGNode::operator=(N);
    FIGNodes = N.FIGNodes;
    Kind = N.Kind;
    return *this;
  }

  FPGNode &operator=(FPGNode &&N) {
    DGNode::operator=(std::move(N));
    FIGNodes = std::move(N.FIGNodes);
    Kind = N.Kind;
    CyclicDep = N.CyclicDep;
    hasReduction = N.hasReduction;
    return *this;
  }

  const FIGNodeSet &getNodes() const { return FIGNodes; }

  NodeKind getKind() const { return Kind; }

  /// Insert the given DDG nodes \p Nodes into this FIG node.
  void insertNodes(const FIGNodeSet &Nodes) {
    FIGNodes.insert(Nodes.begin(), Nodes.end());
  }

  /// Return true if this FIG node contains the given DDG node \p DDGN.
  bool containsNode(const FIGNode &FIGN) const {
    return llvm::find(FIGNodes, &FIGN) != FIGNodes.end();
  }

  /// Populate \p IList with the instructions contained in the DDG nodes
  /// associated with this node. Return true if at least one instructions was
  /// collected, and false otherwise.
  bool collectInstructions(InstructionListType &IList) const;

  /// Populate \p IList with the instructions contained in the DDG nodes
  /// associated with this node, for which the predicate function \p Pred
  /// evaluates to true. Return true if at least one instructions was collected,
  /// and false otherwise.
  bool collectInstructions(InstructionListType &IList,
                           function_ref<bool(Instruction *)> Pred) const;

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FPGNode *N) {
    return N->getKind() == NodeKind::Default;
  }

  bool hasCyclicDep() const { return CyclicDep; }

  void removeUnusedInsts(LoopInfo *LI);

  void fixExternalUses(Loop *L, SmallPtrSet<Instruction *, 4> &Processed);

  void removeDeadLoops(Loop *L, LoopInfo *LI, DominatorTree *DT,
                       ScalarEvolution *SE);

  void setOrigLoop(Loop *L) { OrigLoop = L; }

  Loop *setupClonedLoop(Loop *L, BasicBlock *ExitBlock,
                        BasicBlock *InsertBefore, BasicBlock *LoopDomBB,
                        unsigned Index, LoopInfo *LI, DominatorTree *DT,
                        bool Duplicate) {
    OrigLoop = L;
    if (Duplicate)
      return L;
    MDNode *OrigLoopID = L->getLoopID();
    Loop *NewLoop =
        cloneLoopWithPreheader(InsertBefore, LoopDomBB, Index, LI, DT);
    VMap[ExitBlock] = InsertBefore;
    remapInstructions();
    return NewLoop;
  }

  void remapInstructions() {
    remapInstructionsInBlocks(ClonedLoopBlocks, VMap);
  }

  /// Clones the original loop.
  ///
  /// Updates LoopInfo and DominatorTree using the information that block \p
  /// LoopDomBB dominates the loop.
  Loop *cloneLoopWithPreheader(BasicBlock *InsertBefore, BasicBlock *LoopDomBB,
                               unsigned Index, LoopInfo *LI,
                               DominatorTree *DT) {
    ClonedLoop = llvm::cloneLoopWithPreheader(
        InsertBefore, LoopDomBB, OrigLoop, VMap, Twine(".ldist") + Twine(Index),
        LI, DT, ClonedLoopBlocks);
    return ClonedLoop;
  }

  /// The cloned loop.  If this partition is mapped to the original loop,
  /// this is null.
  const Loop *getClonedLoop() const { return ClonedLoop; }

  /// Returns the loop where this partition ends up after distribution.
  /// If this partition is mapped to the original loop then use the block from
  /// the loop.
  Loop *getDistributedLoop() const {
    return ClonedLoop ? ClonedLoop : OrigLoop;
  }

  /// The VMap that is populated by cloning and then used in
  /// remapinstruction to remap the cloned instructions.
  ValueToValueMapTy &getVMap() { return VMap; }

  void updateStats(Loop *L, const LoopInfo *LI, DominatorTree *DT,
                   ScalarEvolution *SE, InstSet &LoopInsts);
  FPGMetrics &getStats() { return Stats; }

  void printStats() const { Stats.print(); }

  void setReduction() { hasReduction = true; }
  bool isReduction() const { return hasReduction; }
  bool isScalarReduction() const {
    return hasReduction && (Stats.getNumStores() == 0);
  }
  void setCyclicDep(bool value = true) { CyclicDep = value; }

  void dump() const;

  bool handleScalarExpansion(Loop *L, ScalarEvolution *SE,
                             DenseMap<Value *, Value *> &ScalarExpandMap,
                             FissionInterferenceGraph *FIG);

protected:
  FPGNode(const NodeKind Kind) : FPGNodeBase(), Kind(Kind) {}

private:
  FPGNode(FIGNodeSet &Nodes) { FIGNodes.insert(Nodes.begin(), Nodes.end()); }

  /// Data dependence graph nodes associated with this FIG node.
  FIGNodeSet FIGNodes;

  /// The node kind.
  NodeKind Kind = NodeKind::Default;

  SmallPtrSet<Value *, 4> ExtraInsts;

  bool CyclicDep = false;
  bool hasReduction = false;

  Loop *OrigLoop = nullptr;
  Loop *ClonedLoop = nullptr;
  SmallVector<BasicBlock *, 8> ClonedLoopBlocks;
  ValueToValueMapTy VMap;
  FPGMetrics Stats;
};

class FPGEdge : public FPGEdgeBase {
public:
  enum class EdgeKind {
    DefUse, /// edge represents a def-use dependence
    Memory, /// edge represents a data dependence
    Root    /// edge stems from the graph's root node
  };

  FPGEdge(const FPGEdge &) = delete;
  FPGEdge(FPGEdge &&E) : FPGEdgeBase(std::move(E)), Kind(E.getKind()) {}
  virtual ~FPGEdge() = 0;

  /// Create an edge incident to FIG node \p N with the given edge kind \p EK.
  /// When building an affinity edge its weight is given by \p Weight.
  static FPGEdge *create(FPGNode &N, EdgeKind EK);

  FPGEdge &operator=(const FPGEdge &E) {
    FPGEdgeBase::operator=(E);
    Kind = E.Kind;
    return *this;
  }

  FPGEdge &operator=(FPGEdge &&E) {
    FPGEdgeBase::operator=(std::move(E));
    Kind = E.Kind;
    return *this;
  }

  EdgeKind getKind() const { return Kind; }

  void dump() const;

protected:
  FPGEdge(FPGNode &N, EdgeKind Kind) : FPGEdgeBase(N), Kind(Kind) {}

private:
  EdgeKind Kind;
};

class RootFPGNode final : public FPGNode {
public:
  RootFPGNode() : FPGNode(NodeKind::Root) {}
  RootFPGNode(const RootFPGNode &) = delete;
  RootFPGNode(RootFPGNode &&N) : FPGNode(std::move(N)) {}

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FPGNode *N) {
    return N->getKind() == NodeKind::Root;
  }
  static bool classof(const RootFPGNode *N) { return true; }
};

template <FPGEdge::EdgeKind EK> class FPGEdgeOfKind final : public FPGEdge {
public:
  FPGEdgeOfKind(FPGNode &N) : FPGEdge(N, EK) {}
  FPGEdgeOfKind(const FPGEdgeOfKind &) = delete;
  FPGEdgeOfKind(FPGEdgeOfKind &&E) : FPGEdge(std::move(E)) {}

  /// Define classof to be able to use isa<>, cast<>, dyn_cast<>, etc.
  static bool classof(const FPGEdge *N) { return N->getKind() == EK; }
  static bool classof(const FPGEdgeOfKind *) { return true; }
};

using DefUseFPGEdge = FPGEdgeOfKind<FPGEdge::EdgeKind::DefUse>;
using MemoryFPGEdge = FPGEdgeOfKind<FPGEdge::EdgeKind::Memory>;
using RootFPGEdge = FPGEdgeOfKind<FPGEdge::EdgeKind::Root>;

// Fission partition graph contains a set of nodes which represent a
// group of FIGNodes. Each node in this graph represent a partition/Split of
// the original loop. The edges in the graph represent the order in which the
// partitions are dependent on each other.
class FissionPartitionGraph final : public FPGBase {

public:
  using NodeToIndexMapTy = DenseMap<FPGNode *, unsigned>;
  using IndexToNodeMapTy = DenseMap<unsigned, FPGNode *>;
  using NodeToParentMapTy = DenseMap<unsigned, DenseSet<unsigned>>;
  using InstToPartitionIdT = DenseMap<Instruction *, int>;
  using InstSet = SmallPtrSet<Instruction *, 4>;
  using NodeType = FPGNode;
  using NodeListType = SmallVector<NodeType *, 4>;

  FissionPartitionGraph() = delete;
  FissionPartitionGraph(const FissionPartitionGraph &) = delete;
  ~FissionPartitionGraph();

  /// Construct a fission interference graph for the loop nest rooted by \p
  /// OutermostLoop, given the data dependence graph \p DDG.
  static FissionPartitionGraph *create(FissionInterferenceGraph *FIG,
                                       bool AggressiveDistribution,
                                       Loop *OutermostLoop, DominatorTree *DT,
                                       PostDominatorTree *PDT,
                                       ScalarEvolution *SE, LoopInfo *LI);

  bool hasRootNode() const { return (RootNode != nullptr); }
  bool isRootNode(const FPGNode *N) const {
    return (N && isa<RootFPGNode>(*N) && hasRootNode() && *N == getRootNode());
  }

  void populate(bool WithScalarExpansion = true);
  FPGNode &getRootNode() const {
    assert(RootNode && "Root node is not available yet. Graph construction may "
                       "still be in progress\n");
    return *RootNode;
  }

  const DirectedGraph::NodeListTy &getNodes() const { return Nodes; }

  FissionInterferenceGraph *getFIG() const { return FIG; }

  /// Returns the number of nodes in the graph (excluding the root node).
  unsigned size() const {
    return RootNode ? FPGBase::size() - 1 : FPGBase::size();
  }

  /// Retrieve the node \p N graph index.
  unsigned getNodeIndex(const FPGNode &N) const {
    assert(NodeToIndexMap.find(&N) != NodeToIndexMap.end() &&
           "Could not find node in map");
    return NodeToIndexMap.lookup(&N);
  };

  /// Retrieve a graph node given its graph index \p I.
  FPGNode &getNode(unsigned I) const {
    assert(IndexToNodeMap.find(I) != IndexToNodeMap.end() &&
           "Could not find node in map");
    return *IndexToNodeMap.lookup(I);
  };

  void mergeNodes(FPGNode &N1, FPGNode &N2, bool checkStores);

  bool hasSCCs() const;

  bool distributeLoop();

  void dump() const;
  void dumpShort(FPGNode *);
  void dumpShort();

  void viewFPG(const char *suffix = "");

  void printNodeToParentMap() const;

  bool isInvalid() const { return InvalidFPG; }

  bool canMergeExpandedNodes(FPGNode *N1, FPGNode *N2) const;

private:
  FissionPartitionGraph(FissionInterferenceGraph *FIG,
                        bool AggressiveDistribution, Loop *L, DominatorTree *DT,
                        PostDominatorTree *PDT, ScalarEvolution *SE,
                        LoopInfo *LI)
      : FPGBase(), FIG(FIG), AggressiveDistribution(AggressiveDistribution),
        L(L), DT(DT), PDT(PDT), SE(SE), LI(LI) {}

  /// Add the given node \p N to the graph.
  bool addNode(FPGNode &N);

  /// Connect the \p Src and \p Dst nodes using the given edge \p E.
  /// Return true if the edge was added to the graph and false otherwise (an
  /// existing edge was reused).
  bool addEdge(FPGNode &Src, FPGNode &Dst, FPGEdge &E);

  /// Create a root node that connects to every subgraph. For example given:
  ///   G: {N1-->N2,  N3-->N4,  N5}
  /// The root node would be connected to the subgraphs rooted by N1, N3 and N5.
  void createAndConnectRootNode();

  void setupPartitionIdOnInstructions();

  void cloneLoops();

  void setNewLoopID(MDNode *OrigLoopID, FPGNode *Part);

  void removeNodeFromMaps(FPGNode &N) {
    unsigned ix = NodeToIndexMap[&N];
    NodeToIndexMap.erase(&N);
    IndexToNodeMap.erase(ix);
    NodeToParentMap.erase(ix);
  }

  void removeParentFromMap(unsigned ix, unsigned p) {
    NodeToParentMap[ix].erase(p);
  }

  void addParentToMap(unsigned ix, unsigned p) {
    if (ix != p)
      NodeToParentMap[ix].insert(p);
  }

  bool canMergePartition(FPGNode *N1, FPGNode *N2, bool Siblings);
  bool mergeNodeGroup(SmallVectorImpl<FPGNode *> &Nodes, FPGNode *ParentNode,
                      bool Cyclic, bool &ChildNodeMerged,
                      SmallSet<unsigned, 4> &MergedNodes);

  void tryToMergePartitions();

  bool isFPGNodeCyclic(const FPGNode *N) const;

  bool isScalarExpandedNode(const FPGNode *N) const;

  bool areConnected(FPGNode *N1, FPGNode *N2) const;

private:
  FissionInterferenceGraph *FIG;
  bool AggressiveDistribution = false;
  Loop *L = nullptr;

  /// A map from a node to its unique graph index.
  NodeToIndexMapTy NodeToIndexMap;

  /// A map from the node graph index to the node.
  IndexToNodeMapTy IndexToNodeMap;

  NodeToParentMapTy NodeToParentMap;

  InstToPartitionIdT InstToPartitionId;
  /// The graph root node (connected to all subgraphs in the graph).
  RootFPGNode *RootNode = nullptr;

  DominatorTree *DT;
  PostDominatorTree *PDT;
  ScalarEvolution *SE;
  LoopInfo *LI;
  InstSet LoopInsts;

  bool InvalidFPG = false;
  unsigned DotFileCount = 0;
};
raw_ostream &operator<<(raw_ostream &, const FPGNode &);
raw_ostream &operator<<(raw_ostream &, const FPGEdge &);
raw_ostream &operator<<(raw_ostream &, const FissionPartitionGraph &);

//===--------------------------------------------------------------------===//
// GraphTraits specializations for the partition graph
//===--------------------------------------------------------------------===//

/// non-const versions of the grapth trait specializations the FPG.
template <> struct GraphTraits<FPGNode *> {
  using NodeRef = FPGNode *;

  static NodeRef FPGGetTargetNode(DGEdge<FPGNode, FPGEdge> *P) {
    return &P->getTargetNode();
  }

  // Provide a mapped iterator so that the GraphTrait-based implementations can
  // find the target nodes without having to explicitly go through the edges.
  using ChildIteratorType =
      mapped_iterator<FPGNode::iterator, decltype(&FPGGetTargetNode)>;
  using ChildEdgeIteratorType = FPGNode::iterator;

  static NodeRef getEntryNode(NodeRef N) { return N; }
  static ChildIteratorType child_begin(NodeRef N) {
    return ChildIteratorType(N->begin(), &FPGGetTargetNode);
  }
  static ChildIteratorType child_end(NodeRef N) {
    return ChildIteratorType(N->end(), &FPGGetTargetNode);
  }
  static ChildEdgeIteratorType child_edge_begin(NodeRef N) {
    return N->begin();
  }
  static ChildEdgeIteratorType child_edge_end(NodeRef N) { return N->end(); }
};

template <>
struct GraphTraits<FissionPartitionGraph *> : public GraphTraits<FPGNode *> {
  using NodeRef = FPGNode *;
  using nodes_iterator = FissionPartitionGraph::iterator;

  static NodeRef getEntryNode(FissionPartitionGraph *FPG) {
    return &FPG->getRootNode();
  }
  static nodes_iterator nodes_begin(FissionPartitionGraph *FPG) {
    return FPG->findNode(FPG->getRootNode());
  }
  static nodes_iterator nodes_end(FissionPartitionGraph *FPG) {
    return FPG->end();
  }
};

/// const versions of the grapth trait specializations for the FPG.
template <> struct GraphTraits<const FPGNode *> {
  using NodeRef = const FPGNode *;

  static NodeRef FPGGetTargetNode(const DGEdge<FPGNode, FPGEdge> *P) {
    return &P->getTargetNode();
  }

  // Provide a mapped iterator so that the GraphTrait-based implementations can
  // find the target nodes without having to explicitly go through the edges.
  using ChildIteratorType =
      mapped_iterator<FPGNode::const_iterator, decltype(&FPGGetTargetNode)>;
  using ChildEdgeIteratorType = FPGNode::const_iterator;

  static NodeRef getEntryNode(NodeRef N) { return N; }
  static ChildIteratorType child_begin(NodeRef N) {
    return ChildIteratorType(N->begin(), &FPGGetTargetNode);
  }
  static ChildIteratorType child_end(NodeRef N) {
    return ChildIteratorType(N->end(), &FPGGetTargetNode);
  }
  static ChildEdgeIteratorType child_edge_begin(NodeRef N) {
    return N->begin();
  }
  static ChildEdgeIteratorType child_edge_end(NodeRef N) { return N->end(); }
};

template <>
struct GraphTraits<const FissionPartitionGraph *>
    : public GraphTraits<const FPGNode *> {
  using NodeRef = const FPGNode *;
  using nodes_iterator = FissionPartitionGraph::const_iterator;

  static NodeRef getEntryNode(const FissionPartitionGraph *FPG) {
    return &FPG->getRootNode();
  }
  static nodes_iterator nodes_begin(const FissionPartitionGraph *FPG) {
    return FPG->findNode(FPG->getRootNode());
  }
  static nodes_iterator nodes_end(const FissionPartitionGraph *FPG) {
    return FPG->end();
  }
};

} // namespace llvm

#endif // LLVM_TRANSFORMS_SCALAR_FPG_H
