//===- llvm/Analysis/LoopChainPrinter.h -------------------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
//
// This file defines the DOT printer for the Data-Dependence Graph (LoopChain).
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_ANALYSIS_LOOPCHAINPREPRINTER_H
#define LLVM_ANALYSIS_LOOPCHAINPREPRINTER_H

#include "llvm/Transforms/Scalar/LoopChainPre.h"
#include "llvm/Support/DOTGraphTraits.h"


void writePreLCToDotFile(llvm::PreLoopChainGraph &G, bool DOnly);
void writePreLCToDotFile(llvm::PreLoopChainGraph &G, bool DOnly, const int &LCId);

namespace llvm {

//===--------------------------------------------------------------------===//
// Specialization of DOTGraphTraits.
//===--------------------------------------------------------------------===//
template <>
struct DOTGraphTraits<const PreLoopChainGraph *>
    : public DefaultDOTGraphTraits {

  DOTGraphTraits(bool IsSimple = false) : DefaultDOTGraphTraits(IsSimple) {}

  /// Generate a title for the graph in DOT format
  std::string getGraphName(const PreLoopChainGraph *G) {
    assert(G && "expected a valid pointer to the graph.");
    return "PreLCG for '" + std::string(G->getName()) + "'";
  }

  /// Print a LC node either in concise form (-ddg-dot-only) or
  /// verbose mode (-ddg-dot).
  std::string getNodeLabel(const PreLCNode *Node,
                           const PreLoopChainGraph *Graph);

  std::string getNodeAttributes(const PreLCNode *Node,
                                const PreLoopChainGraph *Graph);

  /// Print attributes of an edge in the LC graph. If the edge
  /// is a MemoryDependence edge, then detailed dependence info
  /// available from DependenceAnalysis is displayed.
  
  std::string
  getEdgeAttributes(const PreLCNode *Node,
                    GraphTraits<const PreLCNode *>::ChildIteratorType I,
                    const PreLoopChainGraph *G);
  
  /// Do not print nodes that are part of a pi-block separately. They
  /// will be printed when their containing pi-block is being printed.
  bool isNodeHidden(const PreLCNode *Node, const PreLoopChainGraph *G);

private:
  /// Print a LC node in concise form.
  static std::string getSimpleNodeLabel(const PreLCNode *Node,
                                        const PreLoopChainGraph *G);

  /// Print a LC node with more information including containing instructions
  /// and detailed information about the dependence edges.
  static std::string getVerboseNodeLabel(const PreLCNode *Node,
                                         const PreLoopChainGraph *G);

  /// Print a LC edge in concise form.
  static std::string getSimpleEdgeAttributes(const PreLCNode *Src,
                                             const PreLCEdge *Edge,
                                             const PreLoopChainGraph *G);

  /// Print a LC edge with more information including detailed information
  /// about the dependence edges.
  static std::string getVerboseEdgeAttributes(const PreLCNode *Src,
                                              const PreLCEdge *Edge,
                                              const PreLoopChainGraph *G);
};

using PreLCDotGraphTraits = DOTGraphTraits<const PreLoopChainGraph *>;

} // namespace llvm


#endif // LLVM_ANALYSIS_LCPREPRINTER_H
