//===- StripMining.h - Analyze and strip mine loops --------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//
//===----------------------------------------------------------------------===//
//====  Copyright (c) 2015 Advanced Micro Devices, Inc.  All rights reserved.
//
//               Developed by: Advanced Micro Devices, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// with the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimers.
//
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimers in the documentation
// and/or other materials provided with the distribution.
//
// Neither the names of Advanced Micro Devices, Inc., nor the names of its
// contributors may be used to endorse or promote products derived from this
// Software without specific prior written permission.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH
// THE SOFTWARE.
//
//===----------------------------------------------------------------------===//
//
// This file implements a set of routines that performs strip-mining on loops.
//
//===----------------------------------------------------------------------===//

#include "llvm/Analysis/CGSCCPassManager.h"
#include "llvm/Analysis/LazyCallGraph.h"
#include "llvm/IR/PassManager.h"
#include "llvm/Pass.h"
#include "llvm/ADT/SmallVector.h"
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/DependenceAnalysis.h"
#include "llvm/Analysis/Loads.h"
#include "llvm/Analysis/LoopPass.h"
#include "llvm/Analysis/PostDominators.h"
#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/Analysis/ScalarEvolutionExpressions.h"
#include "llvm/IR/Value.h"
#include "llvm/IR/Verifier.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/Dominators.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Debug.h"
#include "llvm/Analysis/LoopAccessAnalysis.h"
#include "llvm/Analysis/LoopInfo.h"
//#include "llvm/Transforms/Scalar/LoopUnrotate.h"
#include "llvm/Transforms/Utils/AffineTransforms.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/DepUtils.h"
#include "llvm/Transforms/Utils/LoopUtils.h"
#include "llvm/Transforms/Utils/MatrixOperations.h"
#include "llvm/Transforms/Utils/ScalarEvolutionExpander.h"
#include "llvm/Analysis/TargetTransformInfo.h"
#include "llvm/IR/IntrinsicInst.h"
#include "llvm/InitializePasses.h"
#include "llvm/Transforms/Utils/Local.h"
#include "llvm/Transforms/Utils/ValueMapper.h"
#include "llvm/Transforms/Utils/LoopVersioning.h"

using namespace llvm;
using namespace std;
using namespace MatOp;

namespace llvm {

struct StripMining {
  AliasAnalysis *AA;
  LoopInfo *LI;
  DependenceInfo *DA;
  DominatorTree *DT;
  PostDominatorTree *PDT;
  ScalarEvolution *SE;
  OptimizationRemarkEmitter *ORE;
  LoopAccessLegacyAnalysis *LAA = nullptr;
  const TargetTransformInfo *TTI;
  const DataLayout *DL;
  AssumptionCache *AC;
  const LoopAccessInfo *LAI = nullptr;
  std::function<const LoopAccessInfo &(Loop &)> *GetLAA;
  SmallSet<const SCEV*,4> BasePointers;
  SmallSet<Value *,4> DestGEPs;

  bool IRModified = false;

  explicit StripMining(TargetTransformInfo *TTI_, AliasAnalysis *AA_,
            DominatorTree *DT_, PostDominatorTree *PDT_, LoopInfo *LI_,
            AssumptionCache *AC_, ScalarEvolution *SE_, DependenceInfo *DA_,
            std::function<const LoopAccessInfo &(Loop &)> *GetLAA_,
            OptimizationRemarkEmitter *ORE_):
            TTI(TTI_), AA(AA_), DT(DT_), PDT(PDT_), LI(LI_), AC(AC_),
            SE(SE_), DA(DA_), GetLAA(GetLAA_), ORE(ORE_) {}

  // True if Interchange can be done with runtime checks.
  bool withRtChecks = false;

  unsigned LoopTripCount;
  const SCEVAddRecExpr *getARscev(ScalarEvolution *SE, Value *Ptr, Loop *L,
                                  Instruction *LdSt,
                                  SmallVectorImpl<Value *> &MinLBs,
                                  unsigned dim);
  bool checkStripMiningLegality (Loop *L);
  bool CheckConstTripCount(Loop *L);
  bool IsNonContiguousStride(StoreInst *SI, Loop *L);
  bool handleLoopLegalities(Loop *L);
  Loop* StripMineLoop(Loop *L, unsigned SmF);
  std::pair<Value*, Type*> CollectArrayBasePointers(Loop *L);
  bool BufferCandidates(Loop *L) ;
  bool CollectDestGEPs(Loop *L, ArrayType *DestinationArrays,
          AllocaInst* Arrays, PointerType *PointerTy);
  void GenerateBufferingStores(Loop *L, ArrayType *DestinationArrays,
          AllocaInst* Arrays, PHINode* newIdx);
  void StoreFromBufferToDest(Loop *L, ArrayType *DestinationArrays,
          AllocaInst* Arrays, PHINode* newIdx, Type *ElTy,
          PointerType *PointerTy);
  bool CheckBuffering (Loop *L);
  bool IsLegalNontemporalStore(StoreInst *SI, Loop *L,
                                             unsigned int Size) ;
  void setNoAliasToLoop(Loop *VerLoop);
  bool runOnFunction(Function &F);

};

struct StripMiningLegacyPass : public FunctionPass {
  static char ID;
  StripMiningLegacyPass() : FunctionPass(ID) {
    initializeStripMiningLegacyPassPass(*PassRegistry::getPassRegistry());
  }
  bool runOnFunction(Function &F) override;
    void getAnalysisUsage(AnalysisUsage &AU) const override {
    AU.addRequired<AAResultsWrapperPass>();
    AU.addRequired<LoopInfoWrapperPass>();
    AU.addRequired<DependenceAnalysisWrapperPass>();
    AU.addRequired<DominatorTreeWrapperPass>();
    AU.addRequired<PostDominatorTreeWrapperPass>();
    AU.addRequired<ScalarEvolutionWrapperPass>();
    AU.addRequired<LoopAccessLegacyAnalysis>();
    AU.addRequired<OptimizationRemarkEmitterWrapperPass>();
    AU.addRequired<TargetLibraryInfoWrapperPass>();
    AU.addRequired<TargetTransformInfoWrapperPass>();
    AU.addRequired<AssumptionCacheTracker>();
  }
};

struct StripMiningPass: public PassInfoMixin<StripMiningPass> {
  PreservedAnalyses run(Module &M, ModuleAnalysisManager &AM);
  bool runOnFunction(Function &F, FunctionAnalysisManager &AM);
};

} // namespace
